(* eval.fun - evaluation of "is" expressions.				*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1994,1995 by AT&T Bell Laboratories			*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Amy Felty <felty@bell-labs.com>					*)
(*	Elsa Gunter <elsa@research.att.com>				*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)


functor Eval
    (structure Error : ERROR
     structure Lam : LAM
     structure IU : IU
     structure Module : MODULE
     structure Exception : EXCEPTION
     structure CompTable : COMP_TABLE

     val eval_mod : Module.module

     sharing type Lam.constraint = IU.constraint
     sharing type IU.compiled_pred = CompTable.compiled_pred
     sharing type IU.env = Lam.Print.name_env
     sharing type Lam.lterm = Exception.lterm = IU.lterm = CompTable.lterm
     sharing type IU.module = Module.module = CompTable.module) =
struct

local
    open IU Lam
in

    fun unexpected () = raise Basic.InternalError "Eval: unexpected args";

    val eval_table : (const, Print.name_env -> lterm list -> lterm)
	Lib.Table.table =
	Lib.Table.mk_table (Const.hash_const, Lib.Misc.uncurry Const.equal);
	
    val divide_by_zero = get_const eval_mod "divide_by_zero";
    val overflow = get_const eval_mod "overflow";
    val index_out_of_bounds = get_const eval_mod "index_out_of_bounds";
    val not_chr = get_const eval_mod "not_chr";

    fun is_bi M =
	case (LTerm.destruct_const M) of
	    NONE => false
	  | SOME (c, _) =>
		(case (Const.destruct_bi_const c) of
		     NONE => false
		   | _ => true);

    fun eval env M =
	(let
	     val M' = Reduce.head_normalize M
	 in
	     if is_bi M' then
		 M'
	     else
		 let
		     val ((head, _), args) = Reduce.head_args M'
		     val const =
			 case (LTerm.destruct_const head) of
			     SOME (c, _) => c
			   | NONE =>
				 Error.fatal_error
				 (Error.Illegal
				  ("flexible term head on right of is: %s",
				   [Error.STRING
				    (Print.makestring_lterm env M')]))
		 in
		     case (Lib.Table.find eval_table const) of
			 NONE =>
			     Error.fatal_error
			     (Error.Illegal
			      ("Illegal argument to is: %s.",
			       [Error.STRING (Print.makestring_lterm env M')]))
		       | SOME func =>
			     func env args
		 end
	 end) handle
	     Lib.Arith.Div =>
		 raise Exception.exn divide_by_zero
	   | Lib.Arith.Overflow => 
		 raise Exception.exn overflow
	   | Lib.String.Subscript =>
		 raise Exception.exn index_out_of_bounds
	   | Lib.String.Chr =>
		 raise Exception.exn not_chr

    fun bin_oper decompose compose func env [M, N] =
	let
	    val m = decompose (eval env M)
	    val n = decompose (eval env N)
	in
	    compose (func (m, n))
	end
      | bin_oper _ _ _ _ _ =
	raise Basic.InternalError "bin_oper: not binary operation.";

    fun oper decompose compose func env [M] =
	compose (func (decompose (eval env M)))
      | oper _ _ _ _ _ = 
	raise Basic.InternalError "oper: not a unary operation.";

    fun getFoo t the x =
	need (t ^ " arguments to \"is\" funtions must be rigid.")
	(the x);

    val getReal = getFoo "real" theReal;
    val getInt = getFoo "int" theInt;
    val getString = getFoo "string" theString;

    val real_bin_oper	= bin_oper getReal makeReal;
    val real_oper	= oper getReal makeReal;
    val int_bin_oper	= bin_oper getInt makeInt;
    val int_oper	= oper getInt makeInt

    fun lp_substring env [M, N, L] =
	let
	    val m =
		need "string argument to substring must be rigid."
		(theString (eval env M))
	    val n =
		need "int arguments to substring must be rigid."
		(theInt (eval env N))
	    val l =
		need "int arguments to substring must be rigid."
		(theInt (eval env L))
	in
	    makeString (substring (m, n, l))
	end
      | lp_substring _ _ =
	raise Basic.InternalError "lp_substring: bad args.";

(*
    fun lp_explode (Appl (_, M)) =
	let
	    fun makeList (s::ss) =
		Appl (Appl (lp_cons, theString s), makeList ss)
	      | makeList nil =
		lp_nil
	in
	    makeList (explode (theString (eval M)))
	end
      | lp_explode M =
	unexpected M

    fun lp_implode (Appl (_, M)) =
*)

    val opers = [(* binary operations on reals *)
		 ("r+", real_bin_oper (op +)),
		 ("r-", real_bin_oper (op -)),
		 ("r*", real_bin_oper (op * )),
		 ("r/", real_bin_oper (op /)),

		 (* unary operations on reals *)
		 ("r~", real_oper ~),
		 ("rabs", real_oper Lib.Real.abs),
		 ("sqrt", real_oper Lib.Math.sqrt),
		 ("sin", real_oper Lib.Math.sin),
		 ("cos", real_oper Lib.Math.cos),
		 ("arctan", real_oper Lib.Math.atan),
		 ("ln", real_oper Lib.Math.ln),

		 (* functions which convert to/from reals to/from
		  other base types *)
		 ("real_to_string", oper getReal makeString Lib.Convert.ftoa),
		 ("int_to_real", oper getInt makeReal real),
		 ("floor",  oper getReal makeInt Lib.Real.floor),
		 ("ceiling", oper getReal makeInt Lib.Real.ceil),
		 ("round", oper getReal makeInt Lib.Real.round),


		 (* binary operations on integers *)
		 ("+", int_bin_oper (op +)),
		 ("-", int_bin_oper (op -)),
		 ("*", int_bin_oper (op * )),
		 ("div", int_bin_oper (op div)),
		 ("mod", int_bin_oper (op mod)),
		 ("quot", int_bin_oper (Lib.Int.quot)),
		 ("rem", int_bin_oper (Lib.Int.rem)),
		 ("max", int_bin_oper (Lib.Int.max)),
		 ("min", int_bin_oper (Lib.Int.min)),

		 (* unary operations on integers *)
		 ("~", int_oper ~),
		 ("abs", int_oper abs),

		 (* other integer functions *)
		 ("int_to_string", oper getInt makeString Lib.Convert.itoa),

		 ("size", oper getString makeInt size),
		 ("chr", oper getInt makeString Lib.String.chr),
		 ("ord", oper getString makeInt(fn s =>Lib.String.ordof(s,0))),
		 ("^", bin_oper getString makeString (op ^)),
		 ("substring", lp_substring)]


    val _ =
	app (fn (name, func) =>
	     Lib.Table.insert eval_table
	     (case (LTerm.destruct_const (get_const eval_mod name)) of
		  SOME (c, _) => c
		| NONE => raise Basic.InternalError "eval",
	      func))
	opers

    fun is ((_, [lhs, rhs]), IR {env, ...}, con, sc) =
	let
	    val rhs_value = eval env rhs
	    val lhs_evar = Reduce.head_normalize lhs
	in
	    Unify.unify lhs_evar rhs_value con sc
	end 
      | is _ = unexpected ();

    val _ = CompTable.make eval_mod ("is", GFunc is);

end  (* local ... *)
end  (* functor Eval *)


