(* magic.fun - built-in predicates for conversion between strings and	*)
(* terms.								*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)


functor Magic
    (structure ErrorPrint : ERROR_PRINT
     structure Error : ERROR
     structure Program : PROGRAM
     structure Lam : LAM
     structure NameEnv : NAME_ENV
     structure IU : IU
     structure Module : MODULE
     structure Parse : PARSE
     structure Interface : PARSER_INTERFACE
     structure Exception : EXCEPTION
     structure CompTable : COMP_TABLE

     val magic_mod : Module.module

     sharing type Lam.Absyn.aterm = Interface.aterm
     sharing type Lam.Absyn.sign_entry = Module.sign_entry
     sharing type IU.constraint = Lam.constraint
     sharing type CompTable.compiled_pred = IU.compiled_pred
     sharing type CompTable.module = Module.module  = IU.module
     sharing type Program.sterm = Lam.sterm
     sharing type Error.error = ErrorPrint.error
     sharing type Lam.tp = Program.tp = IU.tp
     sharing type Interface.parse_result = Parse.parse_result
     sharing type Lam.Print.name_env = IU.env = Lam.TypeRecon.name_env =
	 NameEnv.name_env
     sharing type IU.lterm = Lam.lterm = Exception.lterm = Program.lterm =
	 CompTable.lterm = NameEnv.evar
     sharing type Lam.const = Module.const = IU.const = Program.const =
	 NameEnv.uvar) =
struct

local
    open Lam IU
in

    fun unexpected () = raise Basic.InternalError "Magic: unexpected args";

    val parse_error = get_const magic_mod "parse_error";
    val type_check_fail = get_const magic_mod "type_check_fail";

    fun term_to_string ((_, [M, N]), IR {env, ...}, con, sc) =
	Unify.unify (makeString (Print.makestring_lterm env M)) N con sc
      | term_to_string _ = unexpected ()

    fun string_to_term ((tp, [str, M]),
			IR {env = name_env, modules, uvars,...}, con, sc) =
	(let
	     fun env cname =
		 (case (Module.sign_lookup modules cname) of
		      res as (SOME _) => res
		    | NONE =>
			  (case (NameEnv.lookup_by_name name_env cname) of
			       SOME (NameEnv.U v) => SOME (Module.Con v)
			     | _ => NONE))			       

	     val (raw_N, location) =
		 case (Parse.parse ("", "") 
		       (Parse.stream_init
			(Lib.IO.open_string
			 (need "string argument to string_to_term must be\
			       \ rigid." (theString str))) "")) of
		     SOME (Interface.ParsedLp (Interface.Term tm), location) =>
			 (Absyn.to_term env tm, location)
		   | _ =>
			 raise Exception.exn
			     (LTerm.construct_application
			      (parse_error,
			       makeString "Illegal interactive declaration"))

	     val tp'' =
		 (case (Type.destruct_arrow tp) of
		      SOME (_, tp') =>
			  (case (Type.destruct_arrow tp') of
			       SOME (tp'', _) => tp''))
		      handle Match =>
			  raise Basic.InternalError "string_to_term"

	     val (evars, N') = Program.freevars_to_evars (raw_N, tp'', uvars)

	     (* We use this function to implement the naming conventions 
	      * described in the file VarProposal.doc located in the doc
	      * directory. This funtion binds evars with names that have
	      * already been used to the corresponding evars in the name
	      * environment. *)
	     fun find_evar ev =
		 case (NameEnv.lookup_by_name name_env
		       ((fn SOME x => x)(LTerm.name_hint ev))) of
		     NONE => ()
		   | SOME (NameEnv.E ev') =>
			 LTerm.bind_evar (ev, ev')

	 in
	     (app find_evar evars;
	      Unify.unify N' M con sc)
	 end
         handle
	 TypeRecon.TypeCheckFail (_, msg) =>
	     raise Exception.exn
		 (LTerm.construct_application
		  (type_check_fail, makeString (msg name_env)))
       |  Error.Error =>
	      let
		  val errors = ref ""
	      in
		  (Error.handle_errors
		   (fn err =>
		    errors := ((ErrorPrint.error_to_string err) ^ "\n" ^
			       (! errors)));
		   raise Exception.exn
		       (LTerm.construct_application
			(parse_error,
			 makeString (! errors))))
	      end)

      | string_to_term _ = unexpected ();
		

    val _ =
	map (CompTable.make magic_mod)
	[("term_to_string", GFunc term_to_string),
	 ("string_to_term", GFunc string_to_term)];

end (* local *)
end; (* functor Magic *)
