(* absyn.fun - functions for constructing simple lambda terms.		*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor Absyn
    (structure Const : CONST
     structure Type : TYPE
     structure Error : ERROR
     structure Fixity : FIXITY
     structure BI : BUILT_INS

     sharing type Type.Const.bi_const = BI.TypeBuiltIns.const
     sharing type Type.tp = Const.tp
     sharing type BI.const = Const.bi_const
     sharing type Const.fixity = Fixity.fixity = Type.Const.fixity)
	 : ABSYN =
struct

    open Const Fixity;
	
    type tp = Type.tp;
    type kind = Type.Kind.kind
    type tp_const = Type.Const.const;
    type const = Const.const;
    type location = Error.location;

    (* generic lambda terms (sterm stands for simple term) - absyn builds
     * sterms which are converted to the lambda terms used by the rest of
     * the code by the LTerm structure. *)
    datatype sterm =

      (* Abstraction *)
	Lam of (tp * string option) * sterm

      (* Application *)
      | App of sterm * sterm

      (* 1-based deBrujin indices - psw fix *)
      | Index of int

      (* Constants (also possibly uvars, introduced by naming mechanism) *)
      | Const of const * tp

      (* Free variables *)
      | Fvar of tp * string

      (* Annotate a term with it location in the source file in order to
       *generate more informative errors.  *)
      | Mark of location * sterm

      (* Explicit type annotations. *)
      | HasType of sterm * tp;

    datatype sign_entry =
	TypeCon of Type.Const.const
      | Con of const;

    (* abstract atom  *)
    type aatom =
	((string -> sign_entry option)* (* lookup function for constants *)
	 string list *			(* bound vars *)
	 (tp * string) list *		(* free vars *)
	 (tp * string) list *		(* wildcard vars *)
	 (string * tp) list)		(* free type vars *)

	-> (tp * string) list *		(* free vars *)
	(tp * string) list *		(* wildcard vars *)
	(string * tp) list *		(* free type vars *)
	sterm atom;			(* atom *)

    (* abstract term *)
    type aterm =
	((string -> sign_entry option)*	(* lookup function for constants *)
	 string list *			(* bound vars *)
	 (tp * string) list *		(* free vars *)
	 (tp * string) list *		(* wildcard vars *)
	 (string * tp) list)		(* free type vars *)

	-> (tp * string) list *		(* free vars *)
	(tp * string) list *		(* wildcard vars *)
	(string * tp) list *		(* free type vars *)
	sterm;				(* simple term *)

    (* abstract term sequence *)
    type atermseq =
	((string -> sign_entry option)*	(* lookup function for constants *)
	 string list *			(* bound vars *)
	 (tp * string) list *		(* free vars *)
	 (tp * string) list *		(* wildcard vars *)
	 (string * tp) list)		(* free type vars *)

	-> (tp * string) list *		(* free vars *)
	(tp * string) list *		(* wildcard vars *)
	(string * tp) list *		(* free type vars *)
	((sterm atom) list);		(* atom list *)

    (* abstract type *)
    type atype =
	((string -> sign_entry option)* (* lookup function for constants *)
	 (string * tp) list)		(* free type vars *)

	-> (string * tp) list *		(* free type vars *)
	tp;				(* type *)

    (* abstract type atom *)
    type atypeatom =
	((string -> sign_entry option)* (* lookup function for constants *)
	 (string * tp) list)		(* free type vars *)

	-> (string * tp) list *		(* free type vars *)
	tp atom;			(* atom *)

    (* abstract type sequence *)
    type atypeseq =
	((string -> sign_entry option)* (* lookup function for constants *)
	 (string * tp) list)		(* free type vars *)

	-> (string * tp) list *		(* free type vars *)
	((tp atom) list);		(* atom list *)
	
    (* abstract kind *)	
    type akind = kind;

    (* identifiers *)
    datatype id' =
      (* Lower-case identifier *)
	Lcid of string
      (* Upper-case identifier *)
      | Ucid of string;

    (* Typed identifier *)
    type id =
	id' * atype;

    fun make_const const =
	Const (const, Type.duplicate_type (type_of const));

    fun to_term env aM =
	let

	    val (freevars, wildvars, _, M) = aM (env, nil, nil, nil, nil)
	in
	    (freevars, wildvars, M)
	end;

    fun to_type env atp =
	let
	    val (_, tp) = atp (env, nil)
	in
	    tp
	end;

    fun to_kind akind = akind;

    (******************** Types ********************)
local
    open  BI.TypeBuiltIns Type Type.Kind;
in

    fun mk_onetypeseq aa location args =
	let
	    val (ftvs, L) = aa args
	in
	    (ftvs, stack construct_application (L, nil))
	    handle FixityError msg =>
		Error.fatal_error
		(Error.Location (location, Error.Illegal (msg, [])))
	end
    
    fun mk_typeseq (aSeq, aL) location (args as (env, _)) =
	let
	    val (ftvs1, seq) = aSeq args
	    val (ftvs2, L) = aL (env, ftvs1)
	in 
	    (ftvs2, stack construct_application (L, seq))
	    handle FixityError msg =>
		Error.fatal_error
		(Error.Location (location, Error.Illegal (msg, [])))
	end;

    fun seq_to_type aSeq location args =
	let
	    val (ftvs, seq) = aSeq args
	in
	    (ftvs, reduce_stack construct_application seq)
	    handle FixityError msg =>
		Error.fatal_error
		(Error.Location (location, Error.Illegal (msg, [])))
	end;

	
    fun type_to_atom atp args =
	let
	    val (ftvs, tp) = atp args
	in
	    (ftvs,
	     case (destruct_const tp) of
		 NONE => Arg tp
	       | SOME c =>
		     (case (Const.fixity c) of
			  Prefix => Arg tp
			| fix as Infix _ =>
			      Opr (tp, Const.makestring_const c, fix)))
	end


    fun type_lookup (_, nil) = NONE
      | type_lookup (vname, (vname', var) :: ls) =
	if vname = vname' then
	    SOME var
	else
	    type_lookup (vname, ls)

    fun mk_ftv vname _ (_, freevars) =
	(case (type_lookup (vname, freevars)) of
	     NONE =>
		 let
		     val var = new_var (SOME vname)
		 in
		     ((vname, var) :: freevars, var)
		 end
	   | SOME var =>
		 (freevars, var))

    fun mk_tconst vname location (env, freevars) =
	(freevars,
	 case (env vname) of
	     SOME (TypeCon const) => construct_const const
	   | _ =>
		 Error.fatal_error
		 (Error.Location (location, Error.UnknownConstant vname)));

    fun mk_tapp (at1, at2) _ (args as (env, _)) =
	let
	    val (fvs1, t1) = at1 args
	    val (fvs2, t2) = at2 (env, fvs1)
	in
	    (fvs2, construct_application (t1, t2))
	end

    fun mk_arrow (at1, at2) _ (args as (env, _)) =
	let
	    val (fvs1, t1) = at1 args
	    val (fvs2, t2) = at2 (env, fvs1)
	in
	    (fvs2, construct_arrow (t1, t2))
	end

    fun mk_one_type (at1, at2) location (args as (env, _)) =
	let
	    val (fvs, t1) = at1 args
	    val (fvs', t2) = at2 (env, fvs)
	in
	    (Type.unify (t1, t2);
	     (fvs', t1))
 	    handle Type.UnifyFailure =>
		Error.fatal_error
		(Error.Location
		 (location,
		  (Error.TypeCheckFail
		   "Idenitfier has differing constraints.")))
	end;

    (* Built-in types *)
    fun mk_bi_type t _ (_, freevars) =
	(freevars, construct_const (Type.Const.construct_bi_const t));
    val mk_out_stream_type = fn x => mk_bi_type OutStreamType x;
    val mk_in_stream_type = fn x => mk_bi_type InStreamType x;
    val mk_string_type = fn x => mk_bi_type StringType x;
    val mk_int_type = fn x => mk_bi_type IntType x;
    val mk_real_type = fn x => mk_bi_type RealType x;


    (* Kinds *) 
    fun mk_type _ =
	Type;

    fun mk_karrow kind _ =
	Karrow Type

end (* local open Type *)

    fun mk_oneseq aa location args =
	let
	    val (fvs, wvs, ftvs, L) = aa args
	in

	    (fvs, wvs, ftvs, stack App (L, nil))
	    handle FixityError msg =>
		Error.fatal_error
		(Error.Location (location, Error.Illegal (msg, [])))
	end
    
    fun mk_termseq (aSeq, aL) location (args as (env, bvars, _, _, _)) =
	let
	    val (fvs1, wvs1, ftvs1, seq) = aSeq args
	    val (fvs2, wvs2, ftvs2, L) = aL (env, bvars, fvs1, wvs1, ftvs1)
	in 
	    (fvs2, wvs2, ftvs2, stack App (L, seq))
	    handle FixityError msg =>
		Error.fatal_error
		(Error.Location (location, Error.Illegal (msg, [])))
	end;

    fun seq_to_term aSeq location args =
	let
	    val (fvs, wvs, ftvs, seq) = aSeq args
	in
	    (fvs, wvs, ftvs, reduce_stack App seq)
	    handle FixityError msg =>
		Error.fatal_error
		(Error.Location (location, Error.Illegal (msg, [])))
	end;
	
    fun find_bvar (vname, bvars) =
	let
	    fun f_b (_, nil) = NONE
	      | f_b (i, name::ls) =
		if vname = name then
		    SOME i
		else
		    f_b (i + 1, ls)
	in
	    f_b (1, bvars)
	end;

    (* make bound variable or free variable *)
    fun mk_bv_fv vname location (_, bvars, freevars, wildvars, freetypevars) =
	(case (find_bvar (vname, bvars)) of
	     SOME i =>
		 (freevars, wildvars, freetypevars,
		  Arg (Mark (location, Index i)))
	   | NONE => 
		 let
		     fun find ((tp, name) :: ls) =
			 if name = vname then SOME tp else find ls
		       | find nil = NONE
		 in
		     case (find freevars) of
			 SOME tp =>
			     (freevars,
			      wildvars,
			      freetypevars,
			      Arg (Mark (location, Fvar (tp, vname))))
		       | NONE =>
			     let
				 val new_tvar = Type.new_var NONE
			     in
				 ((new_tvar, vname) :: freevars,
				  wildvars,
				  freetypevars,
				  Arg (Mark(location, Fvar (new_tvar, vname))))
			     end
		 end);

    fun make_atom (env, location, cname) =
	(case (env cname) of
	     SOME (Con const) =>
		 (case (fixity const) of
		      fix as Infix _ =>
			  Opr (Mark (location, make_const const),
			       makestring_const const,
			       fix)
		    | Prefix =>
			  Arg (Mark (location, make_const const)))
	   | _ =>
		 Error.fatal_error
		 (Error.Location (location, Error.UnknownConstant cname)));

    (* make bound-variable or constant *)
    fun mk_bv_const vname location (env, bvars, freevars,
				    wildvars, freetypevars) =
	(freevars, wildvars, freetypevars,
	 case (find_bvar (vname, bvars)) of
	     SOME i =>
		 Arg (Mark (location, Index i))
	   | NONE =>
		 make_atom (env, location, vname));

    fun mk_const vname location (env, _, freevars, wildvars, freetypevars) =
	(freevars,
	 wildvars, 
	 freetypevars,
	 make_atom (env, location, vname));

    local
	val w_counter = ref 0
    in
	fun mk_wildcard location (_, _, freevars, wildvars, freetypevars) =
	    let
		(* we can name the wildcard with an integer without conflict
		 * because integers are not valid identifiers *)
		val name = Lib.Convert.itoa (Lib.Misc.pre_inc w_counter)
		val binder = (Type.new_var NONE, name)
	    in
		(freevars, binder :: wildvars, freetypevars,
		 Mark (location, Fvar binder))
	    end
    end;

    fun mk_hastype (aM, atp) location (args as (env, _, _, _, _)) =
	let
	    val (fvs, wvs, ftvs1, M) = aM args
	    val (ftvs2, tp) = atp (env, ftvs1)
	in
	    (fvs, wvs, ftvs2, Mark (location, HasType (M, tp)))
	end

    fun atom_to_term aL args =
	let
	    val (fvs, wvs, ftvs, L) = aL args
	    fun bare_atom_to_term (Arg M) = M
	      | bare_atom_to_term (Opr (M, _, _)) = M
        in
	    (fvs, wvs, ftvs, bare_atom_to_term L)
	end;

    fun bare_term_to_atom M =
	let
	    fun opr (HasType (M, _)) = opr M
	      | opr (Mark (_, M)) = opr M
	      | opr (Const (c, _)) =
		(case (fixity c) of
		     fix as Infix _ => SOME (makestring_const c, fix)
		   | _ => NONE)
	      | opr _ = NONE
	in
	    case opr M of
		SOME (name, fix) => Opr (M, name, fix)
	      | NONE => Arg M
	end;
	
    fun term_to_atom aM args =
	let
	    val (fvs, wvs, ftvs, M) = aM args
	in
	    (fvs, wvs, ftvs, bare_term_to_atom M)
	end

    fun id_to_string (Ucid x) = x
      | id_to_string (Lcid x) = x;

    fun id_to_atom (id, atp) location args =
	term_to_atom
	(mk_hastype
	 (atom_to_term
	  (case id of
	       Lcid x => mk_bv_const x location
	     | Ucid x => mk_bv_fv x location),
	  atp)
	 location)
	args;

    fun mk_ucid x _ =
	(Ucid x, fn (_, ftvs) => (ftvs, Type.new_var NONE));

    fun mk_lcid x _ =
	(Lcid x, fn (_, ftvs) => (ftvs, Type.new_var NONE));

    fun mk_of_type ((id, atp), atp') location =
	(id, mk_one_type (atp, atp') location);

    fun mk_abst ((id, atp), aM) location
	(env, bvars, freevars, wildvars, freetypevars) =
	let
	    val vname = id_to_string id
	    val (ftvs', tp) = atp (env, freetypevars)
	    val (fvs, wvs, ftvs'', M) =
		aM (env, vname :: bvars, freevars, wildvars, ftvs')
	in
	    (fvs, wvs, ftvs'', Mark (location, Lam ((tp, SOME vname), M)))
	end;

local
    open BI
in
    fun mk_bi constr x location (_, _, freevars, wildvars, freetypevars) =
	(freevars,
	 wildvars,
	 freetypevars,
	 Mark (location, make_const (construct_bi_const (constr x))));
    val mk_int = fn x => mk_bi Int x;
    val mk_real = fn x => mk_bi Real x;
    val mk_string = fn x => mk_bi String x;
end


end; (* functor Absyn *)
