(* const.fun - constants, and simple operations on them			*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor Const
    (structure BI : BI_CONST
     structure Fixity : FIXITY)
	 : CONST =
struct

    (********************* Types *********************)

    type bi_const = BI.const;
    type tp = BI.tp;

    open Fixity;

    datatype const =
	BI of BI.const

      (* Decl is used for declared constants and Uvars. Constants
       * have stamp 0, while Uvars have higher stamps. The stamp
       * is a reference so that local constants can become Uvars.
       * In addition the stamp ref cell serves as an identifier. *)
      | Decl of
	{name : string option,		(* name *)
	 fixity : fixity ref,		(* fixity and precedence *)
	 defn : const option ref,	(* optional definition (for merging) *)
	 stamp : int ref,		(* time stamp (see above) *)
	 id : int,			(* unique identifier *)
	 tp : tp};			(* type *)

    (********************* Operations on consts *********************)

	
    (******************** creating consts ********************)

local
    val stamp_counter = ref 0;
    val id_counter = ref 0;
    val base_time = ref 0;
in

    (* all constants with time stamps less than or equal to the base_time
     * are constants, while all others are uvars. This may be a bug? *)
    fun reset_base_time () =
	base_time := (! stamp_counter);

    fun is_uvar (Decl {defn = ref (SOME c), ...}) = is_uvar c
      | is_uvar (Decl {stamp = ref i, ...}) = i > (! base_time)
      | is_uvar _ = false;

    fun make_const (tp, name) =
	Decl {tp = tp,
	      name = SOME name,
	      defn = ref NONE,
	      stamp = ref 0,
	      id = Lib.Misc.pre_inc id_counter,
	      fixity = ref Prefix};
	    
    fun new_uvar (tp, name_op) =
	Decl {name = name_op,
	      fixity = ref Prefix,
	      defn = ref NONE,
	      stamp = ref (Lib.Misc.pre_inc stamp_counter),
	      id = Lib.Misc.pre_inc id_counter,
	      tp = tp};

    fun make_local (Decl {defn = ref (SOME c), ...}) =
	make_local c
      | make_local (M as Decl {stamp, ...}) =
	stamp := Lib.Misc.pre_inc stamp_counter;

    (* For time stamping evars. *)
    fun stamp () = ! stamp_counter;

end (* local val id_counter = ... *)


    fun deref (Decl{defn = ref (SOME c), ...}) = deref c
      | deref c = c;

    fun equal (BI bi1) (BI bi2) = BI.equal bi1 bi2
      | equal (Decl {defn = ref (SOME c1), ...}) c2 =
	equal c1 c2
      | equal c1 (Decl {defn = ref (SOME c2), ...}) =
	equal c1 c2
      | equal (Decl {id = id1, ...})
	(Decl {id = id2, ...}) =
	id1 = id2
      | equal _ _ = false;

    fun merge_consts (c1, c2) =
	(* This is important: only merge if the constants are not already
	 * the same. Otherwise a definition loop would be made. *)
	if not (equal c1 c2) then
	    let
		val c1' = deref c1
		fun repoint (Decl {defn = (defn as ref (SOME c)), ...}) =
		    (defn := SOME c1';
		     repoint c)
		  | repoint (Decl {defn, ...}) =
		    defn := SOME c1'
	    in
		repoint c2
	    end
	else
	    ();

    fun hash_const (Decl {defn = ref (SOME c), ...}) = hash_const c
      | hash_const (Decl {id, ...}) = id
      | hash_const (BI bi) = raise Basic.UnImplemented;

    fun name_hint (BI bi) = SOME (BI.makestring_const bi)
      | name_hint (Decl {name, ...}) = name;
	
    fun makestring_const (Decl {name = SOME n, ...}) = n
      | makestring_const (BI bi) = BI.makestring_const bi
      | makestring_const _ =
	raise Basic.InternalError "makestring_const";

    fun type_of (Decl {tp, ...}) = tp
      | type_of (BI bi) = BI.type_of bi
	
    fun age_of (BI bi) = 0
      | age_of (Decl {defn = ref (SOME c), ...}) =
	age_of c
      | age_of (Decl {stamp = ref i, ...}) = i;
	
    fun fixity (BI _) = Prefix
      | fixity (Decl {defn = ref (SOME c), ...}) = fixity c
      | fixity (Decl {fixity = ref fix, ...}) = fix;

    fun update_fixity fix (Decl {defn = ref (SOME c), ...}) =
	update_fixity fix c
      | update_fixity fix (Decl {fixity, ...}) =
	fixity := fix;

    val construct_bi_const = BI;
    fun destruct_bi_const (Decl {defn = ref (SOME c), ...}) =
	destruct_bi_const c
      | destruct_bi_const (BI bi) = SOME bi
      | destruct_bi_const _ = NONE;

    fun is_bi (BI _) = true
      | is_bi _ = false;

end; (* functor Const *)
    
