(* fixity.sml - fixity, associativity, and precedence.			*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

signature FIXITY =
    sig
	datatype associativity =
	    Left
	  | Right
	  | None;

	datatype fixity =
	    Infix of int * associativity
	  | Prefix;

	datatype 'a atom =
	    Opr of 'a * string * fixity
	  | Arg of 'a;

	exception FixityError of string;

	val stack : ('a * 'a -> 'a) -> ('a atom * 'a atom list) ->'a atom list;
	val reduce_stack : ('a * 'a -> 'a) -> 'a atom list -> 'a;
    end; (* signature FIXITY *)

structure Fixity : FIXITY =
struct

    datatype associativity =
	Left
      | Right
      | None;

    datatype fixity =
	Infix of int * associativity
      | Prefix;

    datatype 'a atom =
	Opr of 'a * string * fixity
      | Arg of 'a;
	
    exception FixityError of string;

    fun fixity_of (Opr (_, _, fix)) =
	fix
      | fixity_of _ = raise Basic.InternalError "fixity_of: called on Arg.";

    fun makestring_opr (L as Opr (_, name, fix)) =
	(case fixity_of L of
	     Infix _ => "infix operator "
	   | Prefix  => "prefix operator ")
	     ^ name
      | makestring_opr _ = 
	raise Basic.InternalError "makestring_opr";

    fun term_of (Opr (M, _, _)) = M
      | term_of _ = raise Basic.InternalError "term_of";

    fun reduce_infix app (Arg M2 :: L :: Arg M1 :: seq') =
	Arg(app (app (term_of L, M1), M2)) :: seq'
      | reduce_infix _ _ =
	raise Basic.InternalError "reduce_infix";

    fun reduce_stack _ nil =
	raise Basic.InternalError "reduce_stack: illegal empty stack"

      | reduce_stack _ (Arg M :: nil) = M

      | reduce_stack app (seq as (Arg _ :: (L as Opr _) :: _)) =
	reduce_stack app (reduce_infix app seq)

      | reduce_stack _ (Arg _ :: Arg _ :: _) =
	raise Basic.InternalError "reduce_stack: unreduced application"

      | reduce_stack _ (L :: _) = 
	raise FixityError
	    ("Insufficient arguments for " ^  makestring_opr L)
	    
    fun stackOprArgOpr app (L, Infix (L_prec, None), L'',
			    Infix (L''_prec, None), seq) =
	if L_prec > L''_prec then
	    L :: seq
	else if L_prec = L''_prec then
	    raise FixityError
		("Ambiguous non-associative " ^ makestring_opr L ^
		 " and " ^ makestring_opr L'')
	     else
		 stack app (L, reduce_infix app seq)

      | stackOprArgOpr app (L, Infix (L_prec, _), L'',
			    Infix (L''_prec, Right), seq) =
	if L_prec >= L''_prec then
	    L :: seq
	else
	    stack app (L, reduce_infix app seq)  

      | stackOprArgOpr app (L, Infix (L_prec, _), L'',
			    Infix (L''_prec, _), seq) =
	if L_prec > L''_prec then
	    L :: seq
	else
	    stack app (L, reduce_infix app seq)
      | stackOprArgOpr _ _ = raise Basic.InternalError "stackOprArgOpr"	

    and stack _ (L as Arg _, nil) = L :: nil

      | stack _ (L as Arg _, seq as ((M as Opr _) :: _)) = L :: seq

      | stack app (L as Arg M, (L' as Arg M') :: seq') =
	Arg (app (M', M)) :: seq'

      | stack _ (L as Opr _, nil) =
	raise FixityError ("Leading " ^ makestring_opr L)

      | stack _ (L as Opr _, seq as ((L' as Opr _) :: seq')) =
	raise FixityError
	    (makestring_opr L ^ " following " ^ makestring_opr L')

      | stack _ (L as Opr _, seq as ((Arg _) :: nil)) = L :: seq

      | stack app (L as Opr _, seq as ((Arg _) :: (L' as Opr _) :: _)) =
	stackOprArgOpr app (L, fixity_of L, L', fixity_of L', seq)

      | stack _ (L as Opr _, seq as ((L' as Arg _):: (L'' as Arg _) :: seq')) =
	raise Basic.InternalError "stack: illegally stacked Arg's";
	
end; (* structure Fixity *)
