(* lterm_simple.fun - basic lambda-terms				*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor LTermSimple
    (structure Const : CONST
     structure Evar : VAR
     structure Type : TYPE
     structure Absyn : ABSYN
     structure Trail : TRAIL
     structure PTerm : PTERM

     sharing type Const.associativity = PTerm.associativity
     sharing type Const.const = Absyn.const = Evar.uvar
     sharing type Const.tp = Evar.tp = Type.tp = Absyn.tp)
	 : LTERM_SIMPLE =
struct
	
    structure Type = Type;
    structure Const = Const;
    structure Evar = Evar;

    datatype internal_lterm =

      (* Abstraction - tp holds type of bound variable *)
	Lam of (Type.tp * string option) * internal_lterm

      (* Application *)
      | App of internal_lterm * internal_lterm

      (* Bound variable *)
      | Index of int

      (* Constants and variables introducted by universal quantification
       (also called Uvars) *)
      | ConstUvar of Const.const * Type.tp
	
      (* Logic variable (Existential variable) *)
      | Evar of Evar.var * internal_lterm option ref;

    type lterm = internal_lterm;

    val make_bvar = Index;
    val construct_application = App;
    val construct_abstraction = Lam;
    fun destruct_abstraction (Evar (_, ref (SOME M))) = destruct_abstraction M
      | destruct_abstraction (Lam t) = SOME t
      | destruct_abstraction _ = NONE;
    val construct_const = ConstUvar;
    fun destruct_const (ConstUvar c_tp) = SOME c_tp
      | destruct_const (Evar (_, ref (SOME M))) = destruct_const M
      | destruct_const _ = NONE;

    fun lterm2string (Lam (_, M)) = "Lam " ^ (lterm2string M)
      | lterm2string (App (M, N)) =
	"App (" ^ (lterm2string M) ^ " ," ^ (lterm2string N) ^ ")"
      | lterm2string (Index i) = "Index " ^ (Lib.Convert.itoa i)
      | lterm2string (ConstUvar (cuv, _)) =
	(case (Const.name_hint cuv) of
	     NONE => "noname"
	   | SOME name  => "uvar:"^name)
      | lterm2string (Evar (ev, ref (SOME M))) =
	((case (Evar.name_hint ev) of
	      NONE => "noname-evar"
	    | SOME name => "evar:"^name) ^
	 ".{" ^ (lterm2string M) ^ "}")
      | lterm2string (Evar (ev, _)) =
	(case (Evar.name_hint ev) of
	     NONE => "noname-evar"
	   | SOME name => "evar:"^name)

    val t2s = lterm2string
    fun equal_head (Evar (_, ref (SOME M))) N = equal_head M N
      | equal_head M (Evar (_, ref (SOME N))) = equal_head M N
      | equal_head (Evar (e1, _)) (Evar (e2, _)) = Evar.equal e1 e2
      | equal_head (ConstUvar (c1, _)) (ConstUvar (c2, _)) =
	Const.equal c1 c2
      | equal_head (Index i) (Index j) = i = j
      | equal_head _ _  = false;
	
    fun legal_binding (Evar (evar, _), ConstUvar (uvar, _)) =
	Evar.age_of evar >= Const.age_of uvar 
(*	Lib.List.exists (Const.equal uvar) (Evar.uvars evar)*)
      | legal_binding _ =
	raise Basic.InternalError "legal_binding called on non evar-uvar pair.";

    fun age_of (Evar (_, ref (SOME M))) = age_of M
      | age_of (Evar (evar, _)) = Evar.age_of evar
      | age_of (ConstUvar (cu, _)) = Const.age_of cu
      | age_of _ = raise Basic.InternalError "age_of called on general term";

    fun name_hint (Evar (_, ref (SOME M))) = name_hint M
      | name_hint (Evar (ev, _)) =
	Evar.name_hint ev
      | name_hint (ConstUvar (c_uv, _)) =
	Const.name_hint c_uv
      | name_hint _ = raise Basic.InternalError "name_hint";

    fun older (t1, t2) = age_of t1 < age_of t2;

    (* operations on consts and uvars *)
    fun new_uvar (tp, name_op) =
	ConstUvar (Const.new_uvar (tp, name_op), tp);
	
    (* operations on evars *)
    fun new_evar args = Evar (Evar.new_var args, ref NONE);

    fun bind_evar (Evar (_, t_ref), term) =
	t_ref := SOME term
      | bind_evar _ = raise Basic.InternalError "bind_evar";

    fun unbind_evar (Evar (_, t_ref)) =
	t_ref := NONE
      | unbind_evar _ = raise Basic.InternalError "unbind_evar";

    fun make_evar_as (tp, Evar (_, ref (SOME M))) =
	make_evar_as (tp, M)
      | make_evar_as (tp, Evar (e, _)) =
	Evar
	(Evar.make_var (tp, Evar.age_of e, Evar.name_hint e, Evar.uvars e),
	 ref NONE)
      | make_evar_as _ =
	raise Basic.InternalError "make_evar called on non-evar";

    fun make_evar args = Evar (Evar.make_var args, ref NONE);

    fun muvar uvar =
	ConstUvar (uvar, Const.type_of uvar);

    fun age_of (Evar (_, ref (SOME M))) = age_of M
      | age_of (Evar (evar, _)) = Evar.age_of evar
      | age_of (ConstUvar (uvar, _)) = Const.age_of uvar
      | age_of _ = raise Basic.InternalError "age_of";

    fun type_of (Evar (ev, _)) = Evar.type_of ev
      | type_of (ConstUvar (_, tp)) = tp
      | type_of _ = raise Basic.InternalError "type_of";

    type sterm = Absyn.sterm;

    fun sterm_to_lterm (Absyn.Lam (binder, M)) =
	Lam (binder, sterm_to_lterm M)

      | sterm_to_lterm (Absyn.App (M, N)) =
	App (sterm_to_lterm M, sterm_to_lterm N)

      | sterm_to_lterm (Absyn.Index i) =
	Index i

      | sterm_to_lterm (Absyn.Const c) =
	ConstUvar c

      | sterm_to_lterm (Absyn.Mark (_, M)) = sterm_to_lterm M

      | sterm_to_lterm (Absyn.HasType (M, _)) = sterm_to_lterm M
	 
      | sterm_to_lterm (Absyn.Fvar (tp, str)) =
	Evar (Evar.new_var (tp, SOME str, nil), ref NONE);

    type ('a, 'b) pterm = ('a, 'b) PTerm.pterm;

    fun lterm_to_pterm lt =
	let
	    fun ltp (env, Evar (_, ref (SOME M))) =
		ltp (env, M)
	      | ltp (_, ev as Evar _) =
		PTerm.Evar ev
	      | ltp (_, ConstUvar (c_uv, _)) =
		if Const.is_uvar c_uv then
		     PTerm.Uvar c_uv
		else
		     PTerm.Const c_uv
	      | ltp (env, Index i) =
		(* nth used 0 based indices, our lambda terms use 1 based *)
		(Lib.List.nth (env, i - 1)
		 handle Lib.List.Subscript =>
		     PTerm.Const (Const.make_const (Type.new_var NONE, "_")))

	      | ltp (env, Lam ((tp, name_opt), M)) =
		let
		    val name =
			case name_opt of
			    SOME name => name
			  | NONE => "x"
		    val bvar = Const.new_uvar (tp, SOME name)
		in
		    PTerm.Lam (bvar, ltp ((PTerm.Uvar bvar) :: env, M))
		end
	      | ltp (env, App (App (ConstUvar (const, tp), M), N)) =
		(case (Const.fixity const) of
		     Const.Infix prec_assoc =>
			 PTerm.Oper (const,prec_assoc,ltp(env,M), ltp (env, N))
		   | Const.Prefix =>
			 PTerm.App (PTerm.App (ltp (env, ConstUvar (const,tp)),
					       ltp (env, M)),
				    ltp (env, N)))
		
	      | ltp (env, App (M, N)) =
		PTerm.App (ltp (env, M), ltp (env, N));
	in
	     ltp (nil, lt)
	end;


    fun refresh_types env (Lam ((tp, no), M)) =
	Lam ((Type.copy_type env tp, no), refresh_types env M)
      | refresh_types env (App (M, N)) =
	App (refresh_types env M, refresh_types env N)
      | refresh_types _ (i as Index _) = i
      | refresh_types env (ConstUvar (c, tp)) =
	ConstUvar (c, Type.copy_type env tp)
      | refresh_types env (Evar (_, ref (SOME M))) =
	refresh_types env M
      | refresh_types env (e as Evar _) = e;

    fun instantiate_evar (Evar (_, ref (SOME M)), N) =
	instantiate_evar (M, N)
      | instantiate_evar (Evar (_, r as ref NONE), N) =
	(r := SOME N;
	 Trail.undo (fn () => r := NONE));

end; (* structure LTermSimple *)
