(* print.fun - printing of terms, types, substitutions, and		*)
(* constraints.								*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor PrintTerm
    (structure LTerm : LTERM_SIMPLE
     structure Reduce : REDUCE
     structure Constraints : CONSTRAINTS_TYPES
     structure Formatter : FORMATTER
     structure Switch : SWITCH
     structure Name : NAME_ENV
     structure Unparse : UNPARSE
     structure PTerm : PTERM

     sharing type LTerm.lterm = Name.evar
     sharing type Name.uvar = LTerm.Const.const
     sharing type Name.tvar = LTerm.Type.Var.var
     sharing type LTerm.pterm = LTerm.Type.pterm = PTerm.pterm = Unparse.pterm
     sharing type LTerm.lterm = Constraints.lterm = Reduce.lterm
     sharing type Formatter.format = Unparse.format = LTerm.Type.Kind.format)
	 : PRINT =
struct

    type lterm = LTerm.lterm;
    type tp = LTerm.Type.tp;
    type kind = LTerm.Type.Kind.kind;
    type format = Formatter.format;
    type name_env = Name.name_env;
    type constraint = Constraints.constraint;


local
    open Formatter PTerm;
in
    
    (* Switches *)
    val print_types = ref false;
    val print_abst_types = ref false;
    val _ =
	Switch.make "print_types" (Switch.make_on_off print_types);
    val _ =
	Switch.make "print_abst_types" (Switch.make_on_off print_abst_types);
    (* end of Switches *)


    val makeformat_kind = LTerm.Type.Kind.makeformat_kind;

    fun make_gen_name base =
	let
	    val i = ref 0
	in
	    (fn () =>
	     if !i = 0 then
		 (Lib.Misc.post_inc i;
		  base)
	     else
		 base ^ (Lib.Convert.itoa (Lib.Misc.post_inc i)))
	end;

    fun makeformat_type name_env tp =
	let
	    val pterm = LTerm.Type.type_to_pterm tp

	    fun make_var_name var =
		let
		    val base =
			case (LTerm.Type.Var.name_hint var) of
			    SOME base => base
			  | NONE => "A"
		in
		    Name.add_or_lookup name_env
		    (Name.T var, make_gen_name base)
		end

	    fun format_names (Const const) =
		Const (String (LTerm.Type.Const.makestring_const const))
	      | format_names (Evar var) =
		Evar (String (make_var_name var))
	      | format_names (Uvar _) =
		raise Basic.InternalError
		    "format_names: Uvar found in type pterm."
	      | format_names (Lam (bvar, M)) =
		raise Basic.InternalError
		    "format_names: Lam found in type pterm"
	      | format_names (App (M, N)) =
		App (format_names M, format_names N)
	      | format_names (Oper (oper, prec_assoc, M, N)) =
		Oper (String (LTerm.Type.Const.makestring_const oper),
		      prec_assoc, format_names M, format_names N)
	in
	     Unparse.unparse (format_names pterm)
	end;

    fun makestring_type env tp = makestring_fmt (makeformat_type env tp);

    fun makeformat_lterm name_env lterm =
	let
	    
	    val pterm = LTerm.lterm_to_pterm (Reduce.beta_normalize lterm)

	    fun make_evar_name (ev as LTerm.Evar (evar, _)) =
		let
		    val base =
			case (LTerm.Evar.name_hint evar) of
			    SOME base => base
			  | NONE => "X"
		in
		    Name.add_or_lookup name_env
		    (Name.E ev, make_gen_name base)
		end

	    fun make_uvar_name uvar =
		let
		    val base =
			case (LTerm.Const.name_hint uvar) of
			    SOME base => base
			  | NONE => "u"
		in
		    Name.add_or_lookup name_env (Name.U uvar,
						 make_gen_name base)
		end

	    fun format_names (Const const) =
		Const (String (LTerm.Const.makestring_const const))
	      | format_names (Evar (ev as LTerm.Evar (evar, _))) =
		Evar (HVbox ((String (make_evar_name ev)) ::
			     (if !print_types then
				  [Break, String ":", Break,
				   makeformat_type name_env
				   (LTerm.Evar.type_of evar)]
			      else
				  [])))
	      | format_names (Uvar uvar) =
		Uvar (HVbox ((String (make_uvar_name uvar)) ::
			     (if !print_types then
				  [Break, String ":", Break, 
				   makeformat_type name_env
				   (LTerm.Const.type_of uvar)]
			      else
				  [])))
	      | format_names (Lam (bvar, M)) =
		Lam (HVbox
		     ((String (make_uvar_name bvar)) ::
		      (if !print_types orelse !print_abst_types then
			   [Break, String ":", Break, 
			    makeformat_type name_env(LTerm.Const.type_of bvar)]
		       else
			   [])),
		     format_names M)
	      | format_names (App (M, N)) =
		App (format_names M, format_names N)
	      | format_names (Oper (oper, prec_assoc, M, N)) =
		Oper (String (LTerm.Const.makestring_const oper), prec_assoc,
		      format_names M, format_names N)
		
	in
	    Unparse.unparse (format_names pterm)
	end;

	fun makestring_lterm name_env term =
	    makestring_fmt (makeformat_lterm name_env term);

	(* debugging utilities *)
	val senv  = Name.make_env (fn _ => false);
	val debug = makestring_lterm senv;
	val debug_type = makestring_type senv;
	    

local
    open Constraints LTerm
in
	fun cvt_substitution (M as Evar (_, ref NONE)) =
	    (M, M)

	  | cvt_substitution (Evar (evar, ref (SOME M0))) =
	    (Evar (evar, ref NONE), M0)

	  | cvt_substitution M =
	    raise Basic.InternalError
		"cvt_substitution: not an Evar";

	fun makeformat_subst _ nil =
	    [Space]

	  | makeformat_subst env ((M, N) :: rest) =
	    [Break,
	     HOVbox[makeformat_lterm env M,
		    Space,
		    String "=",
		    Break,
		    makeformat_lterm env N]] @
	    (makeformat_subst env rest);

	fun makestring_substitution env sl = 
	    makestring_fmt
	    (Vbox0 0 1 (makeformat_subst env (map cvt_substitution sl)))

	fun makeformat_dpair env (DPair (eqM, eqN)) =
	    HOVbox [makeformat_lterm env (bare_term eqM),
		    Break,
		    String "=",
		    Space,
		    makeformat_lterm env (bare_term eqN)]

	fun makeformat_dset _ nil =
	    nil

	  | makeformat_dset env (dpair :: rest) =
	    (makeformat_dpair env dpair ::
	     (case rest of
		  nil => []
		| _ => [String ",", Break]))
	    @ makeformat_dset env rest

	fun makeformat_varlist env nil fmts =
	    fmts

	  | makeformat_varlist env (v :: varlist) fmts =
	    makeformat_varlist env varlist
	    (String (makestring_lterm env v)
	     :: String "," :: Break :: fmts)

	fun collect_sorted_vars vars nil =
	    vars

	  | collect_sorted_vars vars (DPair (eqM, eqN) :: rest) =
	    collect_sorted_vars
	    (csv (csv vars (bare_term eqM)) (bare_term eqN))
	    rest

	and csv vars (App (M, N)) =
	    csv (csv vars M) N

	  | csv vars (Lam ((tp, name), M)) =
	    csv vars M

	  | csv vars (Evar (_, ref (SOME M))) =
	    csv vars M

	  | csv vars (M as Evar (_, ref NONE)) =
	    if (Lib.List.exists (equal_head M) vars) then
		vars
	    else
		M :: vars

	  | csv vars (M as ConstUvar (uvar, _)) =
	    if not (Const.is_uvar uvar) orelse
		(Lib.List.exists (equal_head M) vars) then
		vars
	    else
		M :: vars

	  (* Index *)		
	  | csv vars _ = vars;

	fun makeformat_constraint env (Con dset) =
	    let
		val sorted_vars =
		    Lib.List.reverse
		    (Lib.List.sort 
		     (fn (x, y) => age_of x < age_of y)
		     (collect_sorted_vars nil dset))
	    in 
		Hbox [String "(", String "(", Space,
		      Vbox0 0 1
		      (makeformat_varlist env sorted_vars nil @
		       makeformat_dset env dset),
		      Space, String ")", String ")",
		      Newline ()]
	    end

	fun makestring_constraint env con =
	    makestring_fmt (makeformat_constraint env con);

end (* local open Constraints *)

end (* local open LTerm Formatter *)
end; (* functor Print *)
