(* uutils.fun - unification utilities					*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1994,1995 by AT&T Bell Laboratories			*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor UUtils
    (structure LTerm : LTERM_SIMPLE
     structure Reduce : REDUCE
     structure Constraints : CONSTRAINTS_TYPES

     sharing type LTerm.lterm = Reduce.lterm = Constraints.lterm)
	 : UUTILS =
struct

    type lterm = LTerm.lterm;
    type dpair = Constraints.dpair;

    exception NonUnifiable;

local
    open LTerm Constraints;
in

    fun is_flex (App (M, _)) = is_flex M
      | is_flex (Lam (_, M)) = is_flex M
      | is_flex (Evar (_, ref (SOME M))) = is_flex M
      | is_flex (Evar _) = true
      (* Index, ConstUvar *)
      | is_flex _ = false;

    fun abst_over_uvar (uvar as ConstUvar _, term) =
	let
	    fun ao (M as Index _, _) = M
	      | ao (Evar (_, ref (SOME M)), depth) = ao (M, depth)
	      | ao (M as Evar _, _) = M
	      | ao (M as ConstUvar _, d) =
		if equal_head M uvar then
		    Index d
		else
		    M
	      | ao (App (M, N), d) = App (ao (M, d), ao (N, d))
	      | ao (Lam (tp, M), d) =
		Lam (tp, ao (M, d + 1))

	in
	    Lam ((type_of uvar, name_hint uvar), ao (term, 1))
	end

    fun uvars_between (ConstUvar (uvar, _), Evar (evar, _)) =
	let
	    fun se (initseg, nil) = (initseg, nil)
	      | se (initseg, x :: rest) =
		let
		    val initseg' =
			(construct_const (x, Const.type_of x)) :: initseg
		in
		    if Const.equal x uvar then
			(initseg', rest)
		    else
			se (initseg', rest)
		end
	in
	    se (nil, Evar.uvars evar)
	end;

    fun arrow_over_uvar (uvar, tp) =
	Type.construct_arrow (type_of uvar, tp);

    fun abst_over_uvar_raise (uvar as ConstUvar _, term) =
	let
	    fun aor (M as Index _, _) = M
	      | aor (Evar (_, ref (SOME M)), depth) = aor (M, depth)
	      | aor (M as Evar _, d) =
		if legal_binding (M, uvar) then
		    let
			val (w_bar, rest) = uvars_between (uvar, M)
			val raised_type =
			    foldl
			    arrow_over_uvar
			    (type_of M)
			    w_bar
			val raised_evar =
			    make_evar (raised_type,
				       (age_of uvar) - 1,
				       name_hint M,
				       rest)
			val raised_M =
			    foldr
			    (fn (M, N) => App (N, M))
			    raised_evar
			    w_bar
		    in
			(LTerm.instantiate_evar (M, raised_M);
			 aor (raised_M, d))
		    end
		else
		     M
	      | aor (M as ConstUvar _, d) =
		if equal_head M uvar then
		    Index d
		else
		    M
	      | aor (App (M, N), d) =
		App (aor (M, d), aor(N, d))
	      | aor (Lam (tp, M), d) =
		Lam (tp, aor (M, d + 1))

	in
	    (Lam ((type_of uvar, name_hint uvar), aor (term, 1)))
	end;

    fun contained_in_union (l :: ls1, ls2, ls3) =
	(Lib.List.exists (Const.equal l) ls2 orelse
	 Lib.List.exists (Const.equal l) ls3) andalso
	contained_in_union (ls1, ls2, ls3)
      | contained_in_union _ = true;

    fun quick_eoc ((evar as Evar (e, _), args), M) =
	let
	    val u_args = map (fn (ConstUvar (u, _)) => u) args

	    fun qeoc (Evar (_, ref (SOME M))) = qeoc M
	      | qeoc (evar2 as Evar (e2, _)) =
		if (LTerm.older (evar, evar2)) then

		    (* psw this could be optimized! *)
		    (contained_in_union (Evar.uvars e2,
					 Evar.uvars e,
					 u_args))
		else
		     not (equal_head evar evar2)
	      | qeoc (M as ConstUvar (u, _)) =
		legal_binding (evar, M) orelse
		Lib.List.exists (Const.equal u) u_args
	      | qeoc (App (M, N)) = qeoc M andalso qeoc N
	      | qeoc (Lam (_, M)) = qeoc M
	      | qeoc (Index _) = true
	in
	    qeoc M
	end;

    fun head_args_type (M, args) =
	let
	    fun hat (tp, nil) = tp
	      | hat (tp, N :: ls) =
		let
		    val SOME (tp1, tp2) = Type.destruct_arrow tp
		in
		    hat (tp2, ls)
		end
	in
	    hat (type_of M, args)
	end;

    fun extended_occurs_check (gvar as (evar, args), M, dset) =
	let
	    fun eoc (bvars, Evar (_, ref (SOME M)), dset) =
		eoc (bvars, M, dset)

	      | eoc (bvars, M as Evar _, dset) =
		eoc_flex (bvars, M, dset)

	      | eoc (bvars, M as App _, dset) =
		if is_flex M then
		    eoc_flex (bvars, M, dset)
		else
		    eoc_rigid (bvars, M, dset)

	      | eoc (bvars, M as Lam (binder, _), dset) =
		let
		    val a = new_uvar binder
		    val (M', dset') = eoc_norm (a :: bvars, App (M, a), dset)
		in
		    (abst_over_uvar (a, M'), dset')
		end

	      | eoc (bvars, M as ConstUvar _, dset) =
		if legal_binding (evar, M) orelse
		   Lib.List.exists (equal_head M) args orelse
		   Lib.List.exists (equal_head M) bvars then
		   (M, dset)
		else
		    raise NonUnifiable

	      | eoc (_, M, dset) =
		raise Basic.InternalError "extended_occurs_check: loose bvar"
		    
	    and eoc_norm (bvars, M, dset) =
		eoc (bvars, Reduce.head_normalize M, dset)

	    and eoc_flex (bvars, M, dset) =
		if quick_eoc (gvar, M) then (* quick check *)
		    (M, dset)
		else
		    (case (Reduce.head_args M) of
			 ((M' as Evar _, _), args2) =>
			     if equal_head evar M then
				 raise NonUnifiable
			     else
				 let
				     val tp = head_args_type (M', args2)
				 in
				     new_gvar (tp, bvars, M, dset)
				 end
		       | _ => raise Basic.InternalError "eoc_flex: no flex")
			     

	    and eoc_rigid (bvars, App (M, N), dset) =
		let
		    val (M', dset') = eoc_rigid (bvars, M, dset)
		    val (N', dset'') = eoc_norm (bvars, N, dset')
		in
		    (App (M', N'), dset'')
		end

	      (* ConstUvar *)
	      | eoc_rigid (bvars, M, dset) =
		eoc (bvars, M, dset) 

	    and new_gvar (tp, bvars, M, dset) = (* |- M : tp *)
		let
		    val tp' = foldl arrow_over_uvar tp bvars
		    val all_args = args @ rev bvars
		in
		    let
			val tp'' = foldr arrow_over_uvar tp' args 
			val nevar =
			    make_evar_as (tp'', evar)
			val N =
			    foldl (fn (M, N) => App (N, M)) nevar all_args
		    in
			(N, DPair (Gvar (N, (nevar, all_args)), Any M) :: dset)
		    end
		end
	in
	    eoc_norm (nil, M, dset)
	end;

end (* local *)
end; (* functor UUtils *)
