(* type.fun - types							*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor Type
    (structure Const : CONST
     structure Var : VAR
     structure Kind : KIND
     structure Fixity : FIXITY
     structure PTerm : PTERM
     structure Trail : TRAIL

     sharing type PTerm.associativity = Fixity.associativity
     sharing type Const.fixity = Fixity.fixity
     sharing type Const.tp = Kind.kind = Var.tp)
	 : TYPE =
struct
		    
    open Fixity;

    structure Const = Const;
    structure Var = Var;
    structure Kind = Kind;

    datatype tp =
	Arrow of tp * tp
      (* Vars are just an integer stamp. *)
      | Var of Var.var * tp option ref
      | App of tp * tp
      | Const of Const.const;

    (* equality up to a renaming of the type variables. Used only for
     * signature merging. *)
    fun renaming_equal t1 t2 =
	let
	    val env = ref (nil : (Var.var * Var.var) list)
	    fun eq (Arrow (tp1, tp2)) (Arrow (tp3, tp4)) =
		eq tp1 tp3 andalso eq tp2 tp4

	      | eq (App (tp1, tp2)) (App (tp3, tp4)) =
		eq tp1 tp3 andalso eq tp2 tp4

	      | eq (Var (_, ref (SOME tp1))) tp2 =
		eq tp1 tp2

	      | eq tp1 (Var (_, ref (SOME tp2))) = eq tp1 tp2

	      | eq (Var (v1, _)) (Var (v2, _)) =
		(case (Lib.List.lookup (Lib.Misc.uncurry Var.equal) v1 (! env))
		     of
		     NONE =>
			 (env := (v1, v2) :: (! env);
			  true)
		   | SOME v2' =>
			 Var.equal v2 v2')

	      | eq (Const c1) (Const c2) =
		Const.equal c1 c2

	      | eq _ _ = false
	in
	    eq t1 t2
	end;


    fun new_var name_opt =
	Var (Var.new_var (Kind.Type, name_opt, nil), ref NONE);

    val construct_const = Const;
    fun destruct_const (Var (_, ref (SOME t))) = destruct_const t
      | destruct_const (Const c) = SOME c
      | destruct_const _ = NONE;

    val construct_application = App;

    val construct_arrow = Arrow;
    fun destruct_arrow (Var (_, ref (SOME t))) = destruct_arrow t
      | destruct_arrow (Arrow arg) = SOME arg
      | destruct_arrow _ = NONE;

    fun free_in (v, Var (_, ref (SOME t))) = free_in (v, t)
      | free_in (Var (v1, _), Var (v2, _)) = Var.equal v1 v2
      | free_in (v, Arrow (t1, t2)) = free_in (v, t1) orelse free_in (v, t2)
      | free_in (v, App (t1, t2)) = free_in (v, t1) orelse free_in (v, t2)
      | free_in (v, Const _) = false;

    fun result_is_flex (Var (_, ref (SOME t))) = result_is_flex t
      | result_is_flex (Var _) = true
      | result_is_flex (Arrow (_, t)) = result_is_flex t
      | result_is_flex _ = false;

    datatype env = E of (tp * tp) list ref;
    fun make_env () = E (ref (nil : (tp * tp) list));

    fun copy_type (E env) tp =
	let
	    fun dt (Arrow (t1, t2)) =
		Arrow (dt t1, dt t2)
	      | dt (App (t1, t2)) =
		App (dt t1, dt t2)
	      | dt (Var (_, ref (SOME tp))) = dt tp
	      | dt (var as Var (v, _)) =
		(case (Lib.List.lookup
		       (fn (Var (v1, _), Var (v2, _)) => Var.equal v1 v2)
		       var (! env)) of
		     NONE =>
			 let
			     val nv = new_var (Var.name_hint v)
			 in
			     (env := (var, nv) :: (! env);
			      nv)
			 end
		   | SOME nv =>
			 nv)
	      | dt (tp as Const _) = tp
	in
	    dt tp
	end;

    fun duplicate_type tp = copy_type (make_env ()) tp;

    exception UnifyFailure;

    fun inst_var (Var (_, r), tp) =
	(r := SOME tp;
	 Trail.undo (fn () => r := NONE));

    fun unify pair =
	let
	    fun
		(* If all of the term pairs have had variable elimination 
		 * performed on them, then we are done. *)
		unify_aux [] = ()
	      (* *)
	      | unify_aux ((Var (_, ref (SOME t1')), t2) :: ls) =
		unify_aux ((t1', t2) :: ls)
	      | unify_aux ((t1, Var (_, ref (SOME t2'))) :: ls) =
		unify_aux ((t1, t2') :: ls)

	      | unify_aux ((r as Var (v1, _), v as Var (v2, _)) :: ls) =
		(if Var.equal v1 v2 then () else inst_var (r, v);
		 unify_aux ls)

	      | unify_aux ((v as Var _, t) :: ls) =
		if free_in (v, t) then
		    raise UnifyFailure
		else
		    (inst_var (v, t);
		     unify_aux ls)

	      | unify_aux ((t, v as Var _) :: ls) =
		unify_aux ((v, t) :: ls)
 
	      | unify_aux ((Arrow (t1, t2), Arrow (t3, t4)) :: ls) =
		unify_aux ((t1, t3) :: (t2, t4) :: ls)
		
	      | unify_aux ((App (t1, t2), App (t3, t4)) :: ls) =
		unify_aux ((t1, t3) :: (t2, t4) :: ls)

	      | unify_aux ((Const c1, Const c2) :: ls) =
		if Const.equal c1 c2 then
		    unify_aux ls
		else
		    raise UnifyFailure
	      | unify_aux _ = raise UnifyFailure
	in
	    unify_aux [pair]
	end

    fun constructors tp =
	let
	    fun c_aux (pos, Arrow (tp1, tp2)) =
		(c_aux (false, tp1)) @ (c_aux (pos, tp2))
	      | c_aux (pos, App (tp1, tp2)) =
		(c_aux (pos, tp1)) @ (c_aux (pos, tp2))
	      | c_aux (pos, Const c) = [(c, pos)]
	      | c_aux _ = [];
	in
	    c_aux (true, tp)
	end;

    val arrow = Const.make_const (Kind.Type, "->");

    type ('a, 'b) pterm = ('a, 'b) PTerm.pterm;
    fun type_to_pterm (Var (_, ref (SOME t))) =
	type_to_pterm t
      | type_to_pterm (Var (v, _)) = PTerm.Evar v
      | type_to_pterm (Const c) = PTerm.Const c
      | type_to_pterm (App (App (Const c, t1), t2)) =
	(case (Const.fixity c) of
	     Infix prec_assoc =>
		 PTerm.Oper (c, prec_assoc, type_to_pterm t1, type_to_pterm t2)
	   | Prefix =>
		 PTerm.App (PTerm.App (PTerm.Const c, type_to_pterm t1),
			    type_to_pterm t2))	  
      | type_to_pterm (App (t1, t2)) =
	PTerm.App (type_to_pterm t1, type_to_pterm t2)
      | type_to_pterm (Arrow (t1, t2)) =
	PTerm.Oper (arrow, (0, Right), type_to_pterm t1, type_to_pterm t2);

end; (* functor Type *)
