(* type_recon.fun - Type reconstruction					*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor TypeRecon
    (structure Const : CONST
     structure LTerm : LTERM
     structure Location : LOCATION
     structure Print : PRINT
     structure Absyn : ABSYN
     structure Type : TYPE

     sharing type Absyn.location = Location.location
     sharing type Absyn.sterm = LTerm.sterm
     sharing type Print.lterm = LTerm.lterm
     sharing type Absyn.const = Const.const
     sharing type Absyn.tp = Type.tp = Const.tp = Print.tp)
	 : TYPE_RECON =
struct

    exception TypeCheckFail of
    Location.location option * (Print.name_env -> string);

    fun ms_sterm env term =
	Print.makestring_lterm env (LTerm.sterm_to_lterm term);

local
    open Absyn;
in

    type tp = Type.tp;
    type sterm = sterm;
    type location = Location.location;
    type name_env = Print.name_env;

    fun subst (l as Lam (_, t), s) =
	let

	    fun deepen (amount, depth) (Lam (tp, term)) =
		Lam (tp, deepen (amount, depth + 1) term)

	      | deepen a_d (App (t1, t2)) =
		App (deepen a_d t1, deepen a_d t2)

	      | deepen a_d (Mark (loc, M)) =
		Mark (loc, deepen a_d M)

	      | deepen a_d (HasType (M, tp)) =
		HasType (deepen a_d M, tp)

	      | deepen (a, d) (Index i) =
		if i >= d then
		    Index (i + a)
		else
		    Index i
		    
	      (* Const, Fvar *)
	      | deepen _ t = t
		
	    fun do_subst (depth, Index i) =
		if i = depth then
		    deepen (depth - 1, 0) s
		else if depth < i then
		    Index (i - 1)
		     else
			 Index i
			 
	      | do_subst (d, Lam (tp, t))=
		Lam (tp, do_subst (d + 1, t))

	      | do_subst (d, App (t1, t2)) =
		App (do_subst (d, t1), do_subst (d, t2))

	      | do_subst (d, f as Fvar _) =
		f

	      | do_subst (d, Mark (loc, M)) =
		Mark (loc, do_subst (d, M))

	      | do_subst (d, HasType (M, tp)) =
		HasType (do_subst (d, M), tp)

	      | do_subst (_, c as Const _) =
		c
	in
	    do_subst (1, t)
	end
      | subst _ =
	raise Basic.InternalError
	    "first argument to subst is not an abstraction!"

    (* 
     * location tries to skip over implicit arguments to find the location
     * of the head of a term, if available.  This is only a heuristic and
     * there are no guarantees about the returned location!
     *)
    fun left_location (Mark (left, _)) = SOME left
      | left_location (App (M1, M2)) = left_location M1
      | left_location _ = NONE;

    fun right_location (Mark (right, _)) = SOME right
    | right_location (App (M1, M2)) = right_location M2
      | right_location _ = NONE;

    fun location (Mark (location, _)) = SOME location
      | location M =
	(case (left_location M, right_location M) of
	     (NONE, NONE) => NONE
	   | (SOME left, NONE) => SOME left
	   | (NONE, SOME right) => SOME right
	   | (SOME left, SOME right) =>
		 SOME (Location.merge_locations (left, right)));

    fun type_recon (Index _) _ =
	raise Basic.InternalError "type_recon: loose bvar"
	  
      | type_recon (M as Const (_, tp)) sc =
	sc tp M

      | type_recon (M as Fvar (tp, _)) sc =
	sc tp M
      
      | type_recon (App (M1, M2)) sc =
	type_recon M1
	(fn tp => fn M1' =>
	 let
	     fun dest_A tp =
		 case (Type.destruct_arrow tp) of
		     SOME (tp1, tp2) =>
			 type_recon_as_cont tp1 M2
			 (fn M2' => sc tp2 (App (M1', M2')))
		   | NONE => 
			 let
			     val tp' =
				 Type.construct_arrow (Type.new_var NONE,
						       Type.new_var NONE)
			 in
			     (Type.unify (tp, tp');
			      dest_A tp')
			     handle Type.UnifyFailure =>
				 (* Unification failed *)
				 raise TypeCheckFail
				     (location M1,
				      fn env =>
				      (ms_sterm env M1')^" is not a function.")
			 end
	 in
	     dest_A tp
	 end)

      | type_recon (M as (Lam ((tp1, name_opt), M'))) sc =
	let
	    val const = Const.make_const (tp1, "&&&bogus name&&&")

	    fun extract_const (M as Index _, _) = M
	      | extract_const (M as Const (const', _), d) =
		if Const.equal const const' then
		    Index d
		else
		    M
	      | extract_const (App (M, N), d) =
		App (extract_const (M, d), extract_const (N, d))
	      | extract_const (Lam (tp, M), d) =
		Lam (tp, extract_const (M, d + 1))
	      | extract_const (Mark (loc, M), d) =
		Mark (loc, extract_const (M, d))
	      | extract_const (HasType (M, tp), d) =
		HasType (extract_const (M, d), tp)
	      | extract_const (f as Fvar _, _) = f
	in
	    type_recon (subst (M, Const (const, tp1)))
	    (fn tp2 => fn M'' =>
	     sc (Type.construct_arrow (tp1, tp2))
	     (Lam ((tp1, name_opt), extract_const (M'', 1))))
	end

      | type_recon (HasType (M, tp)) sc =
	type_recon_as_cont tp M (fn M' => sc tp M')

      | type_recon (Mark (lrpos, M)) sc =
	type_recon M 
	(fn tp => fn M' =>
	 sc tp (Mark (lrpos, M')))
	
    and type_recon_as_cont tp M sc =
	type_recon M
	(fn tp' => fn M' =>
	 (Type.unify (tp, tp');
	  sc M')
	 handle Type.UnifyFailure =>
	     raise TypeCheckFail
		 (location M,
		  fn env =>
		  (Print.makestring_type env tp) ^
		  " cannot be matched with " ^
		  (Print.makestring_type env tp') ^
		  ", in expression:\n" ^
		  (ms_sterm env M')))

    fun type_recon_as tp M =
	type_recon_as_cont tp M (fn M' => (tp, M'));

end  (* local ... *)
end  (* functor TypeRecon *)
