(* unify.fun - unification a la Huet.					*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Conal Elliott <conal@sun.com>					*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor Unify
    (structure Constraints : CONSTRAINTS_TYPES
     structure LTerm : LTERM
     structure UU : UUTILS
     structure Print : PRINT
     structure Simplify : SIMPLIFY
     structure Trail : TRAIL

     sharing type LTerm.lterm = Constraints.lterm = UU.lterm =
	 Print.lterm
     sharing type Print.tp = LTerm.Type.tp
     sharing type Simplify.constraint = Constraints.constraint)
	 : UNIFY =
struct

    type lterm = LTerm.lterm;
    type constraint = Constraints.constraint;

local
    open Constraints;
in

    val warn_incomplete_ref = ref true;
    val proj_first = ref false;
	
    fun warn_incomplete (head, A) =
	print ("Flexible target type " ^ (Print.debug_type A)
	       ^ " of head " ^ (Print.debug head)
	       ^ " possibly leading to incompleteness.\n")

(*
    fun ms_at (Any t) = "Any (" ^ (Debug.lterm2string t) ^ ")"
      | ms_at (Flex (t, _)) = "Flex (" ^ (Debug.lterm2string t) ^ ")"
      | ms_at (Rigid (t, _)) = "Rigid (" ^ (Debug.lterm2string t) ^ ")"
      | ms_at (Abstraction (t, _)) =
	"Abstraction (" ^ (Debug.lterm2string t) ^ ")"
      | ms_at (Gvar (t, _)) = "Gvar (" ^ (Debug.lterm2string t) ^ ")"

    fun ms_dset (DPair (a1, a2) :: dset) =
	"Dpair (" ^ (ms_at a1) ^ ", " ^ (ms_at a2) ^ ") :: " ^ (ms_dset dset)
      | ms_dset nil = "nil\n"
    and print_dset dset =
	print (ms_dset dset);
*)

    fun type_head_args t  =
	case (LTerm.Type.destruct_arrow t) of
	    SOME (t1, t2) =>
		let
		    val (args, result) = type_head_args t2
		in
		    (t1 :: args, result)
		end
	  | NONE => 
		(nil, t)
			 
    val add_arrows = foldr LTerm.Type.construct_arrow;
    val add_absts = foldr (fn (tp, term) =>
			   LTerm.construct_abstraction ((tp, NONE), term));

    fun MATCH ((_, F), (rigid, (rigid_head, _))) sc =

	let
	    (* |- F : Ftype *)
	    val Ftype = LTerm.type_of F

	    (* |- Ftype == -> Gamma_w . Ftype' *)
	    val (Gamma_w, Ftype') = type_head_args Ftype

	    (* Gamma_w  is  w1:A1,...,wn:An  below *)
	    val n = length Gamma_w

	    (* Now issue a warning if Ftype' is flexible, since the unifier
	     * is incomplete in that case *)
	    val _ =
		if LTerm.Type.result_is_flex Ftype' then
		    warn_incomplete (F, Ftype')
		else
		    ()

	    (* FGamma now needs to be renamed to make the names unique. *)
	    (* val Gamma_w = Sb.rename_vbds FGamma -- to build abstractions *)


	    (* Construct a new Evar Fj with the same characteristics as F
	     * (i.e. of the same age), but of type "PI Gamma_w. B" and apply
	     * to the variables in w1,...,wn in Gamma_w *)
	    fun new_raised_evar B =
		let
		    fun add_args sofar 0 = sofar
		      | add_args sofar i =
			add_args
			(LTerm.construct_application (sofar,
						      LTerm.make_bvar i))
			(i - 1)
		in
		    add_args
		    (LTerm.make_evar_as (add_arrows B Gamma_w, F)) n
		end

	    (* Try a substitution given a term and its type *)

	    (* Given M and A such that  |- M : A instantiate F to M and
	     * constrain  Ftype == A (not necessary in Huet's algorithm) *)
	    fun try_subst_term (M, A) =
		(LTerm.Type.unify (Ftype, A);
		 LTerm.instantiate_evar (F, M);
		 sc ())
		handle LTerm.Type.UnifyFailure => ()

	    (* Add the raised arguments for a given head.  Also return the
	     * type of the constructed term. *)
	    (* |- N : B *)
	    (* For  |- B ==  B1 -> ... -> Bm -> B'
	     * constructs |- N (F1 w1 ... wn) ... (Fn w1 ... wn) : B'
	     * and returns the pair
	     * (N (F1 w1 ... wn) ... (Fn w1 ... wn) , B') *) 
	    fun add_match_rands (N, B) =
		case (LTerm.Type.destruct_arrow B) of
		    SOME (Bj, B') =>
			let
			    val Fj_w1_wn = new_raised_evar Bj
			in
			    add_match_rands
			    (LTerm.construct_application (N,Fj_w1_wn), B')
			end
		  | NONE => (N, B)

	    (* |- head : C *)
	    (* constructs substitution term
	     * |- lam w1:A1 ... wn:An . head (F1 w1 ... wn) ... (Fm w1 ... wn)
	     * : Pi w1:A1 ... wn:An . C' and instantiates F with it *)
	    fun try_head (head, C) =
		let
		    val (head_F1_Fm, C') = add_match_rands (head, C)
		    val _ =
			if LTerm.Type.result_is_flex C' then
			    warn_incomplete (head, C')
			else
			    ()
		in
		    try_subst_term (add_absts head_F1_Fm Gamma_w,
				    add_arrows C' Gamma_w)
		end

	    (* Constructs a substitution term from the rigid head.
	     * This will fail if the rigid head is a Uvar which is not
	     * allowed to appear in the substitution term for F *)
	    (* Tries substitution term ("imitation") of the form
	     * Lam w1:A1 ... wn:An . head (F1 w1 ... wn) ... (Fm w1 ... wn)
	     * where head is the head of the rigid term *)
	    fun imitate () =
		if LTerm.legal_binding (F, rigid_head) then
		    try_head (rigid_head, LTerm.type_of rigid_head)
		     
		else
		    ()			(* F may not depend on a.  Fail *)


	    (* Tries all substitution terms ("projections") of the form
	     * LAM w1:A1 ... wn:An . wi (F1 w1 ... wn) ... (Fm w1 ... wn) *)
	    (* Three cases, since we want to trail only at choice points *)
	    fun project_from (0, nil) = ()
	      | project_from (1, Ai :: nil) =
		try_head (LTerm.make_bvar 1, Ai)
	      | project_from (i, Ai :: Gamma) =
		(Trail.trail (fn () => try_head (LTerm.make_bvar i, Ai));
		 project_from (i - 1, Gamma))

	in
	    if !proj_first then
		(Trail.trail (fn () => project_from (n, Gamma_w)); imitate ())
	    else
		(Trail.trail (fn () => imitate ()); 
		 project_from (n, Gamma_w))
	end;

    fun pick_flex_rigid ((DPair (Flex M, Rigid N)) :: _) =
	SOME (M, N)
      | pick_flex_rigid ((DPair (Rigid N, Flex M)) :: _) =
	SOME (M, N)
      | pick_flex_rigid ((DPair (Gvar (M, (M_head, _)), Rigid N)) :: _) =
	SOME ((M, M_head), N)
      | pick_flex_rigid ((DPair (Rigid N, Gvar (M, (M_head, _)))) :: _) =
	SOME ((M, M_head), N)
      | pick_flex_rigid nil = NONE
      | pick_flex_rigid (_ :: ls) = pick_flex_rigid ls;

    fun unify_dset dset sc =
	(let
	     val Con dset' = Simplify.simplify_constraint (Con dset)
	 in
	     case (pick_flex_rigid dset') of
		 NONE =>
		     sc (Con dset')
	       | SOME flex_rigid =>
		     MATCH flex_rigid (fn () => unify_dset dset' sc)
	 end)
	     handle Simplify.NonUnifiable => ()
		     
    fun unify M N (Con dset) sc =
	unify_dset ((DPair (Any M, Any N)) :: dset) sc;

    val is_flex = UU.is_flex;

end; (* local *)

end; (* functor Unify *)
