(* error_print.fun - functor for structure which converts errors to	*)
(* printable error messages.						*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor ErrorPrint
    (structure Error : ERROR
     structure TerminalIO : TERMINAL_IO
     structure Formatter : FORMATTER
     structure Location : LOCATION

     sharing type Location.location = Error.location)
	 : ERROR_PRINT =
struct

    type error = Error.error;

    structure F = Formatter;

local
    open Error
in
  
    exception BadFormat of string;

    fun format_error (format_string, format_list) =
	let
	    fun fe ("%"::"%"::ls) fs =
		(F.String "%") :: (fe ls fs)
	    
	      | fe [] [] = []
	    
	      | fe ("%"::c::ls) (f::fs) =
		(F.String
		 (case (f, c) of
		      (STRING string, "s") => string
		    | (INT n, "i") => (Lib.Convert.itoa n)
		    | (REAL r, "r") => (Lib.Convert.ftoa r)
		    | (BOOL b, "b") => (Lib.Convert.btoa b)
		    | _ => raise
			  (BadFormat
			   ("Type mismatch between escape and object: " ^
			    format_string))))
		:: (fe ls fs)
		  
	      | fe ("%"::_) [] =
		raise (BadFormat ("More escapes than objects: " ^
				  format_string))
		
	      | fe (c :: cs) fs =
		(if Lib.String.is_whitespace c then
		     F.Break
		 else
		     (F.String c))
		:: (fe cs fs)
	    
	      | fe [] (_::_) =
		raise (BadFormat ("More objects than escapes: " ^
				  format_string))
		
	in
	    fe (Lib.String.explode format_string) format_list
	end;

    fun makestring_loc location =
	(Location.makestring_location location) ^ ":"

    fun e2s (loc_opt, UnknownConstant cname) =
	(loc_opt,
	 [F.String "Undeclared constant", F.Break,F.String cname,F.String "."])

      | e2s (loc_opt, UnknownModule module) =
	(loc_opt,
	 [F.String "Unknown module", F.Break, F.String module, F.String "."])

      | e2s (loc_opt, UnknownSwitch switch) =
	(loc_opt,
	 [F.String "Unknown switch", F.Break, F.String switch, F.String "."])

      | e2s (_, Location (loc, err)) =
	e2s (SOME loc, err)

      | e2s (loc_opt, ParseError err) =
	e2s (loc_opt, err)

      | e2s (loc_opt, Parse err) =
	(loc_opt, format_error err)

      | e2s (loc_opt, TypeCheckFail str) =
	(loc_opt,
	 [F.String "Type checking failed:", F.Break, F.String str])

      | e2s (loc_opt, Illegal err) =
	(loc_opt, format_error err);

    fun error_to_string err = 
	case (e2s (NONE, err)) of
	    (SOME loc, str) =>
		F.makestring_fmt
		(F.HVbox
		 [F.String (makestring_loc loc), F.Break, F.String "Error:",
		  F.Break, F.HVbox str])
	  | (NONE, str) =>
		F.makestring_fmt
		(F.HVbox [F.String "Error:", F.Break, F.HVbox str]);

    fun error_print err =
	TerminalIO.print_error (error_to_string err);

    fun handle_errors () =
	Error.handle_errors error_print;

end (* local open Error *)
end; (* functor ErrorPrint *)
