(* interpreter.fun - functor which contains code that implements the	*)
(* main loop of the query engine, and all logical primitives.		*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1994,1995 by AT&T Bell Laboratories			*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)
(* Also, see "A semi-functional implementation of a higher-order logic	*)
(* programming language" by Conal Elliott and Frank Pfenning, in Peter	*)
(* Lee, editor, Topics in Advanced Language Implementation, pages	*)
(* 289-325, MIT Press, 1991.						*)

(* Authors:								*)
(*	Amy Felty <felty@bell-labs.com>					*)
(*	Elsa Gunter <elsa@research.att.com>				*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)


(* Version of INTERPRETER for extended Horn clauses
   including side-effect with trailing
   and higher-order terms *)

functor Interpreter
   (structure Error : ERROR
    structure Lam : LAM
    structure IU : IU
    structure TerminalIO : TERMINAL_IO
    structure Lg : LOGICALS
    structure Module : MODULE
    structure Program : PROGRAM 
    structure CompTable : COMP_TABLE

    sharing type Lam.Print.name_env = IU.env
    sharing type CompTable.compiled_pred = IU.compiled_pred
    sharing type Lam.const = Lg.const = Module.const = CompTable.const =
	IU.const
    sharing type Module.module = IU.module
    sharing type IU.constraint = Lam.constraint
    sharing type Program.clause = IU.clause = Module.clause
    sharing type IU.tp = Lam.tp = Module.tp = CompTable.tp
    sharing type IU.lterm = Program.lterm = CompTable.lterm = Lam.lterm =
	Module.lterm)
	: INTERPRETER =
struct

    exception Quit;

    type lterm = Lam.lterm;
    type constraint = Lam.constraint;
    type i_record = IU.i_record;

local

    open Lam IU;

    fun trace (M, con, env, t :: ts) =
	(Trail.trail
	 (fn () =>
	  (Unify.unify t M con
	   (fn _ => TerminalIO.message ("Solving goal: " ^
					(Print.makestring_lterm env M)))));
	 trace (M, con, env, ts))
      | trace (_, _, _, nil) = ()


    exception Unexpected;

    (* Logical Constants *)
    fun truth (_, _, con, sc) = sc con

    and conjunction ((_, [g1, g2]), record, con, sc) =
	solve (g1, record, con, fn newcon => solve (g2, record, newcon, sc))
      | conjunction _ = raise Unexpected
      
    and disjunction ((_, [g1, g2]), record, con, sc) =
	(Trail.trail (fn () => solve (g1, record, con, sc));
	 solve (g2, record, con, sc))
      | disjunction _ = raise Unexpected

    and implication ((_, [d, g]), record, con, sc) =
	add_clause (d, g, record, con, sc)
      | implication _ = raise Unexpected

    and back_implication ((hd, [g, d]), record, con, sc) =
	implication ((hd, [d, g]), record, con, sc)
      | back_implication _ = raise Unexpected
      
    and uquant ((tp, [g]), record as IR {uvars, ...}, con, sc) =
	let
	    (* we use theAbstraction to get the binder of the abstraction,
	     * so that we can use the bound variable name as the name hint
	     * for uvar, if possible *)
	    val (binder, _) = theAbstraction (arg_arg_type tp) g
	    val uvar = LTerm.new_uvar binder
	    val SOME (u, _) = LTerm.destruct_const uvar
      in
	  solve (LTerm.construct_application (g, uvar),
		 add_uvars (u :: uvars, record), con, sc)
      end
    | uquant _ = raise Unexpected
	  
    and equant ((tp, [g]), record as IR {uvars, ...}, con, sc) =
	let
	    (* Again we use theAbstraction to retain the name of
	     * bound variable, if possible. *)
	    val ((tp, n_opt), _) = theAbstraction (arg_arg_type tp) g
	    val evar = LTerm.new_evar (tp, n_opt, uvars)
	in
	    solve (LTerm.construct_application (g, evar), record, con, sc)
	end
      | equant _ = raise Unexpected
      
  (* end of Logical Constants *)

    and add_clause (d, g, record as (IR {clauses, ...}), con, sc) =
	let
	    val dclauses =
		Program.clausify d clauses
	in
	    solve (g, add_clauses (dclauses, record), con, sc)
	end

    and solve (M, record as (IR {traced, env, ...}), con, sc) =
	let
	    val M' = Reduce.head_normalize M
	    val _ = trace (M', con, env, traced)
	    val ((head, tp), args) =
		if Unify.is_flex M' then
		    Error.fatal_error
		    (Error.Illegal
		     ("Term head %s is not rigid.",
		      [Error.STRING (Print.makestring_lterm env M)]))
		else
		    Reduce.head_args M'
	    val head_const =
		case (LTerm.destruct_const head) of
		    SOME (c, _) => c
		  | NONE =>
			raise Basic.InternalError "solve: term head is Index." 
	in
	    case (CompTable.find head_const) of
		SOME (GFunc func) =>
		    (func ((tp, args), record, con, sc)
		     handle Unexpected =>
			 unexpected M')
	      | NONE =>
		    solve_atom (M', head_const, record, con, sc)
	end
  
    and solve_atom (M, h, record, con, sc) =
	let
	    (* The ctag (cut tag?) is used to implement the cut (!)
	     * predicate. *)
	    exception new_ctag
	in
	    match_atom (M, h, add_ctag (new_ctag, record), con, sc)
	    handle new_ctag => ()
	end

    and match_atom (M, M_head,
		    record as (IR {modules,clauses,env,uvars, ...}), con, sc) =
	let
	    fun rec_match nil = ()
	      | rec_match ((clause as Module.Clause (N_head, _, _, _))::rest) =
		if Const.equal N_head M_head then
		    (Trail.trail (fn () => try_clause false nil clause);
		     rec_match rest)
		else
		    rec_match rest

	    and global_lookup () =
		Module.prog_lookup modules M_head (try_clause true)

	    and try_clause refresh imports (Module.Clause (L, vars, N, gbody))=
		let
		    (* The tenv is used to "refresh" the types in the clause.
		     * This is done because, as the code is now written, we
		     * have no way to quantify over the type variables at the
		     * term level.  This is a bug. *)
		    val tenv = Type.make_env ()

		  val evars =
		      map
		      (fn (tp, no) =>
		       LTerm.new_evar
		       (if refresh then Type.copy_type tenv tp else tp,
			no,
			uvars))
		      vars

		  fun make_term M =
		      foldr
		      (fn (evar, M) =>
		       LTerm.construct_application (M, evar))
		      M evars

		  val N' = make_term (if refresh then
					  LTerm.refresh_types tenv N
				      else
					  N)
	      in
		  Unify.unify N' M con
		  (fn con' =>
		   (let
			val mods = imports @ modules
			val gbody' =
			    make_term (if refresh then 
					   LTerm.refresh_types tenv gbody
				       else
					   gbody)
			val n_uvars =
			    Lib.List.flatten (map Module.inst_locals imports)
			val uvars' = n_uvars @ uvars
			val record' =
			    add_modules (mods, add_uvars (uvars', record))
		    in
			solve (gbody', record', con', sc)
		    end))
	      end
	in 
	    (rec_match clauses;
	     global_lookup ())
	end

    val logicals =
	[(Lg.ttrue, truth),
	 (Lg.comma, conjunction),
	 (Lg.ampersand, conjunction),
	 (Lg.semicolon, disjunction),
	 (Lg.arrow, implication),
	 (Lg.backarrow, back_implication),
	 (Lg.pi, uquant),
	 (Lg.sigma, equant)];
    val _ = app (fn (c, f) => CompTable.store (c, GFunc f)) logicals;
	
in

    val solve = solve;
end  (* local ... *)

end  (* functor Interpreter *)
