(* interpreter_utils.fun - *)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor InterpreterUtils
    (structure Error : ERROR
     structure Name : NAME_ENV
     structure Lam : LAM
     structure Module : MODULE

     sharing type Module.const = Lam.const)
	 : IU =
struct

local
    open Lam
in

    type lterm = Lam.lterm;
    type const = Const.const;
    type tp = Lam.tp;
    type bi_const = BI.const;
    type module = Module.module;
    type clause = Module.clause;
    type env = Name.name_env;
    type constraint = Lam.constraint;
	

    datatype i_record =
	IR of
	{modules : module list,
	 clauses : clause list,
	 env : env,
	 uvars : const list,
	 ctag : exn,
	 traced : lterm list};
     
    (* The type of built-in predicates, including logical constants *)
    datatype compiled_pred =
	GFunc of
	(tp *				(* type of head of arg *)
	 lterm list) *			(* args *)
	i_record *			(* record *)
	constraint *			(* con *)
	(constraint -> unit)		(* sc *)
	-> unit;

    fun some (SOME x) = x;

    fun theFoo destruct M =
	SOME
	(destruct
	 (some (Const.destruct_bi_const
		(#1 (some (LTerm.destruct_const M))))))
	handle Match => NONE;
	    

    fun need _ (SOME x) = x
      | need str NONE =
	Error.fatal_error (Error.Illegal (str, nil));

    val theInt = theFoo (fn BI.Int i => i);
    val theReal = theFoo (fn BI.Real r => r);
    val theString = theFoo (fn BI.String s => s);
    val theInStream =
	theFoo (fn (BI.InStream (stream, _)) => stream);
    val theOutStream =
	theFoo (fn (BI.OutStream (stream, _)) => stream);

    (*
     * theAbstraction takes a type t1 and a term of type t1 -> t2
     * and makes an explicit abtraction, eta-expanding, if
     * necessary. theAbstraction returns the binder of the abstraction
     * and the body of the abstraction.
     *)
    fun theAbstraction tp M =
	case (LTerm.destruct_abstraction M) of
	    SOME res =>
		res
	  | NONE =>
		((tp, NONE),
		 LTerm.construct_application (M, LTerm.make_bvar 1));

    fun get_const module name =
	case (Module.sign_lookup [module] name) of
	    SOME (Module.Con exn) =>
		LTerm.construct_const (exn, Const.type_of exn)
	  | _ => 
		(print ("Error: get_const can't find " ^ name ^ "\n");
		 raise Basic.InternalError "get_const.");

    fun makeBi bi =
	let
	    val c = (Const.construct_bi_const bi)
	in
	    LTerm.construct_const (c, Const.type_of c)
	end;

    val makeString = makeBi o BI.String;
    val makeReal = makeBi o BI.Real;
    val makeInt = makeBi o BI.Int;
	
    fun add_traced (new_traced,
		    IR {modules, clauses, env, ctag, traced, uvars}) =
	IR {modules = modules, clauses = clauses, env = env, uvars = uvars,
	    ctag = ctag, traced = new_traced};

    fun add_ctag (new_ctag, IR {modules, clauses, env, ctag, traced, uvars}) =
	IR {modules = modules, clauses = clauses, env = env,
	    ctag = new_ctag, traced = traced, uvars = uvars};

    fun add_clauses (new_clauses,
		     IR {modules, clauses, env, ctag, traced, uvars}) =
	IR {modules = modules, clauses = new_clauses, env = env,
	    ctag = ctag, traced = traced, uvars = uvars};

    fun add_modules (new_modules,
		     IR {modules, clauses, env, ctag, traced, uvars}) =
	IR {modules = new_modules, clauses = clauses, env = env,
	    ctag = ctag, traced = traced, uvars = uvars};

    fun add_uvars (new_uvars, IR{modules, clauses, env, ctag, traced, uvars}) =
	IR {modules = modules, clauses = clauses, env = env,
	    ctag = ctag, traced = traced, uvars = new_uvars};

    fun unexpected M =
	raise Basic.InternalError "Unexpected form of goal.";

    fun bin_pred decompose pred =
	GFunc
	(fn ((_, [M, N]), _, con, sc) =>
	 if pred (decompose M, decompose N) then sc con else ()
          | _ => 
		raise Basic.InternalError "bin_pred: not a binary predicate.");

    fun arg_arg_type tp =
	case (Type.destruct_arrow tp) of
	    SOME (tp', _) =>
		(case (Type.destruct_arrow tp') of
		     SOME (tp'', _) => tp''
		   | NONE =>
			 raise Basic.InternalError "arg_arg_type: 1")
	  | NONE => 
		raise Basic.InternalError "arg_arg_type: 2";

end (* local open Lam *)
end; (* functor InterpreterUtils *)
