(* link.sml - application of functors to form the structures of which	*)
(* Terzo is composed.							*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1994,1995 by AT&T Bell Laboratories			*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Amy Felty <felty@bell-labs.com>					*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)


structure TerminalIO = TerminalIO ();

structure Module : MODULE =
    Module (structure Lam = Lam
	    structure Formatter = Formatter
	    structure Name = NameEnv
	    structure TerminalIO = TerminalIO
	    structure Error = Error);

fun get_mod name =
    case (Module.lookup_module name) of
	SOME m => m
      | NONE => raise Basic.InternalError "get_mod";

structure SetUp : SET_UP =
    struct
	type module = Module.module;

	val pervasive_mods : module list ref = ref [];
	val lam_home = Lib.Dir.cwd () ^ "/"
	val path = ["/pkg/terzo/lib"];
    end;

structure I = 
struct
    val pervasive_mods = SetUp.pervasive_mods;

    exception Error of (int * int) * Error.error;
    datatype parserarg =
	A of {name : string,
	      counter : LexerInterface.counter};
end;

structure ParserInterface : PARSER_INTERFACE =
    ParserInterface (structure Absyn = Absyn
		     structure Fixity = Fixity);

structure LPLrVals : LP_LRVALS =
    LPLrValsFun (structure Token = LrParser.Token
		 structure Module = Module
		 structure LexerInterface = LexerInterface
		 structure ParserInterface = ParserInterface
		 structure Error = Error
		 structure Absyn = Absyn
		 open I);
    
structure LPLex : ARG_LEXER =
    LPLexFun (structure Error = Error
	      structure TerminalIO = TerminalIO
	      structure Tokens = LPLrVals.Tokens
	      structure LexerInterface = LexerInterface);

structure LPParser : ARG_PARSER =
   JoinWithArg (structure ParserData = LPLrVals.ParserData
		structure Lex = LPLex
		structure LrParser = LrParser);

structure ParseCore : PARSE =
    ParseCore (structure ParserInterface = ParserInterface
	       structure Location = LexerInterface
	       structure TerminalIO = TerminalIO
	       structure Module = Module
	       structure Absyn = Absyn
	       structure Error = Error
	       structure Parser = LPParser
	       open I
	       val EOF_token_name = LPLrVals.Tokens.EOF
	       val EOP_token_name = LPLrVals.Tokens.DOT);

structure ErrorPrint : ERROR_PRINT =
    ErrorPrint (structure Error = Error
		structure TerminalIO = TerminalIO
		structure Formatter = Formatter
		structure Location = LexerInterface);


structure SignMerge : SIGN_MERGE =
    SignMerge (structure Error = Error
	       structure Lam = Lam
	       structure TerminalIO = TerminalIO
	       structure Formatter = Formatter
	       structure Switch = Switch
	       structure Module = Module);
    

structure ParseSign : PARSE_MODULE =
    ParseModule (structure Error = Error
		 structure ErrorPrint = ErrorPrint
		 structure Lam = Lam
		 structure Interface = ParserInterface
		 structure Parse = ParseCore
		 structure Module = Module
		 structure Switch = Switch
		 structure TerminalIO = TerminalIO
		 structure Program =
		     struct
			 type sterm = Lam.sterm;
			 type lterm = Lam.lterm;
			 type tp = Lam.tp;
			 type const = Lam.const;
			 type clause = Module.clause;
			 exception NotDForm of lterm;
			 
			 fun bogus _ = raise Basic.UnImplemented;
			 val abst_over_fvar = bogus;
			 val freevars_to_uquant = bogus;
			 val freevars_to_evars = bogus;
			 val clausify = bogus;
			 val terms_to_clauses = bogus;
		     end;
		 structure SignMerge = SignMerge
		 structure SetUp = SetUp);

(**************************************************************)
(* Bootstrapping logical pervasives                           *)
(**************************************************************)
(* Read the declarations of the logical pervasives necessary  *)
(* understand clauses. This side-effects the Module structure *)

structure Dummy1 =
struct

    val _ = print "Reading signature entries for logical primitives.\n";
    val (close_func, stream) =
	ParseCore.file_init (Lib.Dir.mk_path (SetUp.lam_home,
					      "lp-lib/bi/logicals.mod"))
    val _ =
	(ParseSign.parse_module stream;
	 close_func ();
	 print "Done.\n")
	handle Error.Error =>
	    (print "\nError while loading Logicals\n\n";
	     Error.handle_errors ErrorPrint.error_print)

    val lg = get_mod "Logicals";
    val _ = SetUp.pervasive_mods := [lg];

end;

structure Lg : LOGICALS =
    Logicals (structure Module = Module
	      structure Lam = Lam
	      val lg = Dummy1.lg);


structure IU : IU =
    InterpreterUtils (structure Module = Module
		      structure Name = NameEnv
		      structure Error = Error
		      structure Lam = Lam);

structure Program : PROGRAM =
    Program (structure TypeRecon = TypeRecon
	     structure Lam = Lam
	     structure Lg = Lg
	     structure Error = Error
	     structure Module = Module
	     structure UU = UUtils
	     structure IU = IU);

structure ParseModule : PARSE_MODULE =
    ParseModule (structure Error = Error
		 structure ErrorPrint = ErrorPrint
		 structure Lam = Lam
		 structure Interface = ParserInterface
		 structure Parse = ParseCore
		 structure Module = Module
		 structure Switch = Switch
		 structure TerminalIO = TerminalIO
		 structure Program = Program
		 structure SignMerge = SignMerge
		 structure SetUp = SetUp);

structure CompTable : COMP_TABLE =
    CompTable (structure Lam = Lam
	       structure Module = Module
	       structure IU = IU);
    

structure Interpreter : INTERPRETER =
    Interpreter (structure Error = Error
		 structure Lam = Lam
		 structure IU = IU
		 structure TerminalIO = TerminalIO
		 structure Lg = Lg
		 structure Module = Module
		 structure Program = Program
		 structure CompTable = CompTable);

structure Exn =
    struct
	exception exn of Lam.lterm;
    end;

structure Query : QUERY =
    Query (structure Switch = Switch
	   structure Error = Error
	   structure ErrorPrint = ErrorPrint
	   structure Lam = Lam
	   structure Interface = ParserInterface
	   structure Parse = ParseCore
	   structure Lg = Lg
	   structure Name = NameEnv
	   structure Module = Module
	   structure IU = IU
	   structure Program = Program
	   structure Interpreter = Interpreter
	   structure TerminalIO = TerminalIO
	   structure SetUp = SetUp
	   open Exn);

structure Help : HELP =
    Help (structure Error = Error
	  structure Formatter = Formatter
	  structure TerminalIO = TerminalIO);

structure Script : SCRIPT =
    Script (structure Error = Error
	    structure ErrorPrint = ErrorPrint
	    structure TerminalIO = TerminalIO
	    structure Module = Module
	    structure SignMerge = SignMerge
	    structure Absyn = Absyn
	    structure Query = Query
	    structure ParseModule = ParseModule
	    structure Interface = ParserInterface
	    structure Parse = ParseCore
	    structure Switch = Switch
	    structure Help = Help
	    val boot_path = [Lib.Dir.mk_path (SetUp.lam_home, "lp-lib/bi/")]
	    val pervasive_mods = SetUp.pervasive_mods)

(**************************************************************)
(* Installing the pervasive signature and program             *)
(**************************************************************)

structure Dummy2 =
struct
    val _ = print "Reading first boot script.\n";
    val (close_func, stream) =
	ParseCore.file_init (Lib.Dir.mk_path (SetUp.lam_home,
					      "lp-lib/bi/stage1.boot"))
    val _ =
	Script.read_script stream
	handle Error.Error =>
	    (ErrorPrint.handle_errors ();
	     raise Error.Error)

    val _ = (close_func (); print "Done.\n")
end;


(*
 * The following two structures are not used in the building of other
 * structures. They are included simply for the side effect that create
 * in the CompTable structure.
 *)

structure Exception : EXCEPTION =
    Exception (structure Lam = Lam
	       structure Interpreter = Interpreter
	       structure IU = IU
	       structure CompTable = CompTable
	       val exn_mod = get_mod "Exception"
	       open Exn
	       val _ = print "Exception\n");

structure StreamIo = 
    StreamIo (structure Unify = Unify
	      structure Error = Error
	      structure CompTable = CompTable
	      structure IU = IU
	      structure Lam = Lam 
	      structure Module = Module
	      val stream_io_mod = get_mod "StreamIo"
	      val _ = print "Stream IO\n");

structure File =
    File (structure Unify = Unify
	  structure Exception = Exception
	  structure Error = Error
	  structure IU = IU
	  structure Lam = Lam
	  structure Module = Module
	  structure CompTable = CompTable
	  val file_mod = get_mod "File"
	  val _ = print "File\n");


structure Magic =
    Magic (structure Unify = Unify
	   structure ErrorPrint = ErrorPrint
	   structure Error = Error
	   structure Lam = Lam
	   structure NameEnv = NameEnv
	   structure IU = IU
	   structure Module = Module 
	   structure Exception = Exception
	   structure Parse = ParseCore
	   structure Interface = ParserInterface
	   structure Program = Program
	   structure CompTable = CompTable
	   val magic_mod = get_mod "Magic"
	   val _ = print "Magic\n");
	       
structure Eval = 
    Eval (structure Error = Error
	  structure Lam = Lam
	  structure IU = IU
	  structure Module = Module
	  structure Exception = Exception
	  structure CompTable = CompTable
	  val eval_mod = get_mod "Eval"
	  val _ = print "Eval\n");


structure LpString =
    LpString (structure CompTable = CompTable
	      structure IU = IU
	      val string_module = get_mod "String"
	      val _ = print "String\n");

structure LpReal =
    LpReal (structure CompTable = CompTable
	    structure IU = IU
	    val real_module = get_mod "Real"
	    val _ = print "Real\n");


structure LpInt =
    LpInt (structure CompTable = CompTable
	   structure IU = IU
	   val int_module = get_mod "Int"
	   val _ = print "Int\n");

structure LpList : LP_LIST =
    LpList (structure LTerm = LTerm
	    structure Module = Module
	    val list_mod = get_mod "List");

structure Control =
    Control (structure IU = IU
	     structure CompTable = CompTable
	     structure I = Interpreter
	     val control_mod = get_mod "Control"
	     val _ = print "Control\n");

structure Trace =
    Trace (structure IU = IU
	   structure CompTable = CompTable
	   structure Interpreter = Interpreter
	   val trace_mod = get_mod "Trace"
	   val _ = print "Trace\n");

(*
structure Ground =
    Ground (structure IU = IU
	    structure CompTable = CompTable
	    structure Module = Module
	    structure LpList = LpList
	    structure Unify = Unify
	    structure LTerm = LTerm
	    val ground_mod = get_mod "Ground"
	    val _ = print "Ground\n");
*)

(* Read the System module. This needs to be done after the bi structures
 * are compiled because some of them side-effect the signatures which will
 * be imported in System. *)
structure Dummy3 =
struct
    val _ = print "Reading second boot script.\n";
    val (close_func, stream) =
	ParseCore.file_init (Lib.Dir.mk_path (SetUp.lam_home,
					      "lp-lib/bi/stage2.boot"))
    val _ =
	Script.read_script stream
	handle Error.Error =>
	    (ErrorPrint.handle_errors ();
	     raise Error.Error)
	    
    val _ = (close_func (); print "Done.\n")
    val _ = SetUp.pervasive_mods := [get_mod "System"];
end;

structure Top : TOP =
    Top (structure Error = Error
         structure TerminalIO = TerminalIO
	 structure Switch = Switch
	 structure Parse = ParseCore
	 structure ErrorPrint = ErrorPrint
	 structure Script = Script);



fun save () = SMLofNJ.exportFn (SetUp.lam_home ^ "../bin/.lpsml", Top.top);
fun run () = Top.top ("", ["TERZO_PATH=" ^
			   (Lib.Dir.mk_path (SetUp.lam_home, "lp-lib/"))]);
