(* module.fun - module manipulation					*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)


functor Module
    (structure Lam : LAM
     structure Formatter : FORMATTER
     structure Name : NAME_ENV
     structure TerminalIO : TERMINAL_IO
     structure Error : ERROR

     sharing type Lam.Print.kind = Lam.Type.Const.tp
     sharing type Lam.Print.name_env = Name.name_env
     sharing type Formatter.format = Lam.Print.format)
	 : MODULE =
struct

    type tp = Lam.tp;
    type lterm = Lam.lterm;
    type const = Lam.const;
    type tp_const = Lam.Type.Const.const;

    open Lam.Absyn; (* for sign_entry constructors *)
	
local
    open Lam
in
    datatype clause = Clause of
	const *				(* head of head of implication *)
	(tp * string option) list *	(* free vars *)
	lterm *				(* head of implication *)
	lterm;				(* body of implication *)
	
structure Sign =
struct
    type sign = (string, sign_entry) Lib.Table.table;

    fun make_sign () =
	Lib.Table.mk_table (Lib.String.hash_string, op =);

    fun insert sign pair =
	Lib.Table.insert sign pair;

    fun lookup sign name =
	Lib.Table.find sign name;

    fun make_local sign names =
	app (Lib.Table.remove sign) names;

    val make_localkind = make_local;

    fun accumulate sign signs =
	app (fn s => (Lib.Table.map (insert sign) s; ())) signs
	
    val import = accumulate;

    (* To make SignMerge.sign_merge easy to write, we ensure that all
     * type constructors are apped first. *)
    fun sign_app sign func =
	let
	    (* This ensures that all TypeCons are listed first *)
	    fun separate ((tcs, cs), (se as (_, TypeCon _)) :: ls) =
		separate ((se :: tcs, cs), ls)
	      | separate ((tcs, cs), (se as (_, Con _)) :: ls) =
		separate ((tcs, se :: cs), ls)
	      | separate ((tcs, cs), nil) = tcs @ cs
	in
	    app func (separate ((nil, nil), Lib.Table.list_items sign))
	end;

end; (* structure Sign *)


structure Prog =
struct

    local
	val id_counter = ref 0;
    in
	fun new_id () =
	    Lib.Misc.pre_inc id_counter;
    end

    datatype inc =
	ACC
      | IMP;

    datatype prog =
	P of {id : int,
	      prog_tab : (const, clause list) Lib.Table.table,
	      included : (inc * prog) list,
	      locals : const list};

    fun make_prog_tab () =
	Lib.Table.mk_table (Const.hash_const,
			fn (x, y) => Const.equal x y);

    fun make_prog () =
	P {id = new_id (),
	   prog_tab = make_prog_tab (),
	   included = nil,
	   locals = nil};

    (* We need to be careful that the sorted groups that sort_clauses
     * returns retain any initial clause ording (That is, sort_clause
     * must be a stable sort), because the ording indicates the clause
     * selection order.  *)
    fun sort_clauses clauses =
	let
	    val clause_groups : (const * clause list ref) list ref =
		ref []

	    fun find (clause as Clause (M_head, _, _, _)) =
		let
		    fun f [] = NONE
		      | f ((N_head, N_clauses)::ls) =
			if Const.equal M_head N_head then
			    SOME N_clauses
			else
			    f ls
		in
		    f (! clause_groups)
		end

	    fun store (clause as Clause (M_head, _, _, _)) =
		case (find clause) of
		    NONE => clause_groups := (M_head, ref [clause]) ::
			(! clause_groups)
		  | SOME ls => ls := (! ls) @ [clause]
	in
	    (app store clauses;
	     map (fn (head, ref cls) => (head, cls)) (! clause_groups))
	end;

    fun insert (P {included, locals, ...}) clause_list =
	let
	    val clause_groups = sort_clauses clause_list
	    val prog_tab = make_prog_tab ()
	    val _ = app (Lib.Table.insert prog_tab) clause_groups
	in
	    P {id = new_id (), included = included, locals = locals,
	       prog_tab = prog_tab}
	end;

    (* The sc should have the type module list -> clause -> unit,
     * where the module list is the list of imports. *)
    fun lookup module extract modules const sc =
	let
	    val progs = map extract modules

	    (* Table to store the names of modules which have already 
	     * been searched, so that we don't do repeat lookups. *)
	    val table : (int, unit) Lib.Table.table =
		Lib.Table.mk_table (fn x => x, op =);

	    fun module_lookup (P {id, included, prog_tab, ...})=
		if (Lib.Table.find table id = SOME ()) then
		    ()
		else
		    (Lib.Table.insert table (id, ());
		     let
			 val accumulated =
			     Lib.List.filter 
			     (fn (IMP, _) => NONE
			   | (ACC, module) => SOME module)
			     included
			 val sc' = sc (map (module o #2) included)
		     in
			 (app module_lookup accumulated;
			  case (Lib.Table.find prog_tab const) of
			      NONE => ()
			    | SOME clause_list => 
				  app 
				  (fn clause => 
				   (Trail.trail 
				    (fn () => sc' clause)))
				  clause_list)
		     end)
	in
	    app module_lookup progs
	end;

    fun list (P {prog_tab, ...}) =
	Lib.Table.list_items prog_tab;

    fun inst_locals (P {locals, ...}) =
	(app Const.make_local locals;
	 locals)

    fun make_local (P {locals, included, prog_tab, ...}) consts =
	(P {locals = consts @ locals, included = included, prog_tab = prog_tab,
	    id = new_id ()});

    fun empty (P {included, prog_tab, ...}) =
	null included andalso null (Lib.Table.list_items prog_tab);

    fun inc inc_type (P {locals, included, prog_tab, ...}) progs =
	let
	    (* This is a small optimization: empty programs are
	     * not included. *)
	    val progs' =
		Lib.List.filter
		(fn prog =>
		 if empty prog then NONE else SOME (inc_type, prog))
		progs
	in
	    (P {locals = locals,
		included = included @ progs',
		prog_tab = prog_tab,
		id = new_id ()})
	end;

    val accumulate = inc ACC;
    val import = inc IMP;
	
end; (* structure Prog *)

   open Sign Prog;
 
   datatype module =
	M of {name : string,
	      sign : sign,
	      prog : prog ref};

    val module_table : (string, module) Lib.Table.table = 
	Lib.Table.mk_table (Lib.String.hash_string, op =);

    fun new_module module_name =
	let
	    val new_mod = M {name = module_name,
			     sign = make_sign (),
			     prog = ref (make_prog ())}
	in
	    (Lib.Table.insert module_table (module_name, new_mod);
	     new_mod)
	end;

    fun remove_module module_name =
	Lib.Table.remove module_table module_name;

    fun lookup_module module_name =
	Lib.Table.find module_table module_name;

    fun module_name (M {name, ...}) = name;

    fun sign_lookup ((M {sign, name, ...}) :: modules) cname =
	(case (Sign.lookup sign cname) of
	     res as (SOME _) => res
	   | NONE => sign_lookup modules cname)
      | sign_lookup nil name =
	NONE;

    fun is_const modules name =
	case (sign_lookup modules name) of
	    NONE => false
	  | SOME (TypeCon _) => false
	  | _ => true;

    fun prog_lookup modules const sc =
	Prog.lookup
	(fn prog => M {name = "&&&bogus&&&",
		       sign = make_sign (),
		       prog = ref prog})
	(fn (M {prog = ref p, name, ...}) => p)
	modules	const sc;

    fun sign_insert (M {sign, ...}) pair =
	Sign.insert sign pair;

    fun prog_insert (M {prog, ...}) pair =
	prog := Prog.insert (!prog) pair;

    fun prog_list (M {prog, ...}) = Prog.list (! prog);

    fun inst_locals (M {prog, ...}) =
	Prog.inst_locals (! prog);

    fun make_local (M {prog, sign, ...}) names =
	let
	    val consts =
		map (fn x =>
		     (case (Sign.lookup sign x) of
			  SOME (Con c) => c
			| SOME _ =>
			      raise Basic.InternalError
				  ("make_local: " ^ x ^ " is typecon.")
			| NONE => 
			      raise Basic.InternalError
				  ("make_local: " ^ x ^ " does not exist.")))
		names
	in
	    (prog := Prog.make_local (! prog) consts;
	     Sign.make_local sign names)
	end;

    fun make_localkind (M {sign, ...}) names =
	Sign.make_localkind sign names;

    fun accumulate (M {sign, prog, ...}) modules =
	(prog :=
	 Prog.accumulate (! prog) (map (fn (M {prog=ref p,...}) =>p) modules);
	 Sign.accumulate sign (map (fn (M {sign, ...}) => sign) modules));

    fun import (M {sign, prog, ...}) modules =
	(prog :=
	 Prog.import (! prog) (map (fn (M {prog=ref p,...}) =>p) modules);
	 Sign.import sign (map (fn (M {sign, ...}) => sign) modules));

local
    open Formatter
in
    fun fmt_entry (tk, makeformat_tp, name, fixity) const =
	Vbox0 0 1 
	((HVbox
	  [String tk, Break,
	   String name, Break,
	   makeformat_tp const, String "."]) ::
	
	 (case (fixity const) of
	      Const.Prefix => []
	    | Const.Infix (prec, assoc) =>
		  [Break,
		   HVbox
		   [String ("infix" ^
			    (case assoc of
				 Const.Left => "l"
			       | Const.Right => "r"
			       | Const.None => "")), Break,
		    String name, Break,
		    String (Lib.Convert.itoa prec), String "."]]));

end (* local open Formatter *)

    fun makeformat_entry sign (name, Con const) =
	let
	    val env =
		Name.make_env (fn str =>
			       case (Sign.lookup sign str) of
				   SOME _ => true
				 | _ => false)
	in
	    fmt_entry ("type", (Print.makeformat_type env) o Const.type_of,
		       name, Const.fixity) const
	end
      | makeformat_entry _ (name, TypeCon tc) =
	fmt_entry ("kind", Print.makeformat_kind o Type.Const.type_of,
		   name, Type.Const.fixity) tc;

    fun print_se sign (name, se) =
	TerminalIO.print
	((Formatter.makestring_fmt (makeformat_entry sign (name, se))) ^ "\n");

    fun print_sign (M {sign, ...}) =
	let
	    val entries : (string * sign_entry) list ref = ref nil
	    val _ = Sign.sign_app sign (fn x => entries := x :: (! entries))
	    val entries' =
		Lib.List.sort
		(fn ((x, _), (y, _)) => Lib.String.lessthan (y, x))
		(! entries)
	in
	    app (print_se sign) entries'
	end;

    fun sign_app (M {sign, ...}) func =
	Sign.sign_app sign func;

end (* local open Lam *)
end; (* functor Module *)
