(* parse_module.fun - reads zero or more modules from an input stream.	*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor ParseModule
    (structure Error : ERROR
     structure ErrorPrint : ERROR_PRINT
     structure Lam : LAM
     structure Parse : PARSE
     structure Interface : PARSER_INTERFACE
     structure Module : MODULE
     structure Switch : SWITCH
     structure TerminalIO : TERMINAL_IO
     structure Program : PROGRAM
     structure SignMerge : SIGN_MERGE
     structure SetUp : SET_UP
     sharing type Lam.Const.fixity = Interface.fixity
     sharing type Interface.akind = Lam.Absyn.akind
     sharing type Interface.atype = Lam.Absyn.atype
     sharing type Interface.aterm = Lam.Absyn.aterm
     sharing type Module.sign_entry = Lam.Absyn.sign_entry
     sharing type Lam.Print.name_env = Lam.TypeRecon.name_env
     sharing type Lam.lterm = Program.lterm
     sharing type Error.error = ErrorPrint.error
     sharing type Module.const = Lam.const
     sharing type Interface.parse_result = Parse.parse_result
     sharing type Lam.sterm = Program.sterm
     sharing type Error.location = Parse.location = Lam.TypeRecon.location
     sharing type SignMerge.module = Module.module = SetUp.module
     sharing type Module.clause = Program.clause
     sharing type Module.tp_const = Lam.Type.Const.const = SignMerge.tp_const
     sharing type Lam.tp = Program.tp)
	 : PARSE_MODULE =
struct

    (* switches *)
    val print_sigs = ref false;
    val _ = Switch.make "print_sigs" (Switch.make_on_off print_sigs);
    (* end of switches *)


local
    open Lam Error Interface;
in

    type parse_stream = Parse.parse_stream;

    (*
     * val update_fixities : module -> (fixity * int) * string list -> unit 
     *)
    fun update_fixities module (fix_prec, names) =
	let
	    fun update_fixity token =
		(case (Module.sign_lookup [module] token) of
		     NONE =>
			 ()
		   | SOME (Module.TypeCon tconst) =>
			 Type.Const.update_fixity fix_prec tconst
		   | SOME (Module.Con const) => 
			 Const.update_fixity fix_prec const)
	in
	    app update_fixity names
	end;
		 
    fun store_sign_entries cstr make_const module (A, cnames) =
	let
	    fun store_sign_entry cname =
		((case (Module.sign_lookup [module] cname) of
		      SOME _ =>
			  error
			  (Illegal ("Redefinition of constant %s.",
				    [STRING cname]))
		    | NONE => ());
		 let
		     val const = make_const (A, cname)
		 in
		     (Module.sign_insert module (cname, cstr const);
		      (cname, const))
		 end)
	in
	    map store_sign_entry cnames
	end;

    val store_types =
	store_sign_entries Module.TypeCon Type.Const.make_const;
    val store_consts =
	store_sign_entries Module.Con Const.make_const;

    fun parse_module_decl stream =
	case (Parse.parse ("| ", "| ") stream) of
	    SOME (ParsedModule (Module module_name), _) =>
		SOME module_name

	  | SOME (ParsedScript End, location) =>
		NONE
	    
	  | NONE =>
		NONE

	  (* The rest of the cases handle errors. *)
	  | SOME (ParsedModule _, _) =>
		fatal_error
		(Illegal
		 ("Expected module declaration, found preamble declaration.",
		  []))

	  | SOME (ParsedLp _, _) =>
		fatal_error
		(Illegal
		 ("Expected module declaration, found lp declaration.",
		  []))

	  | SOME (ParsedScript _, _) =>
		fatal_error
		(Illegal
		 ("Expected module declaration, found script declaration.",
		  []));

    datatype inc =
	Imp of Module.module
      | Acc of Module.module;

    fun get_module location cstr name =
	case (Module.lookup_module name) of
	    (SOME module)  => SOME (cstr module)
	  | NONE => (error (Location (location,UnknownModule name));
		     NONE);
	    
    fun parse_preamble (parse, unparse) =
	case (parse ()) of
	    SOME (ParsedModule (Preamble (Import imp_names)), location) =>
		(Lib.List.filter (get_module location Imp) imp_names) @
		(parse_preamble (parse, unparse))
		
	  | SOME (ParsedModule (Preamble (Accumulate acc_names)), location) =>
		(Lib.List.filter (get_module location Acc) acc_names) @
		(parse_preamble (parse, unparse))

	  | other =>
		(unparse other;
		 nil);

    fun parse_body (parse, unparse) =
	let
	    fun pb (SOME (ParsedLp (Sign entries), location))
		(sign_entries, terms) =
		pb (parse ()) (sign_entries @ [(entries, location)], terms)
		
	      | pb  (SOME (ParsedLp (Term vars_M), location)) (se, t) =
		pb (parse ()) (se, (vars_M, location) :: t)
		 
	      | pb (SOME (ParsedScript End, _)) lls =
		(lls, false)

	      | pb NONE lls =
		(lls, false)
	    
	      | pb (mod_dec as (SOME (ParsedModule (Module _), _))) lls =
		(unparse mod_dec;
		 (lls, true))

	      | pb _ _ =
		fatal_error
		(Illegal
		 ("Illegal module declaration. Missing #end statement?", []))
	in
	    pb (parse ()) ([], [])
	end;

    fun parse_module stream =
	pm (parse_module_decl stream, stream)

    and pm (NONE, _) = Basic.success
      | pm (SOME module_name, stream) =
	let
	    fun parse () =
		Parse.parse (module_name ^ "|\t", module_name ^ "|\t") stream
	    fun unparse arg =
		Parse.unparse stream arg
	    val pair = (parse, unparse)

	    (* Parse a module. All parsing errors will occur during the
	     * next two statements. This does not include fixity and 
	     * unknown constant errors. *)
	    val include_list = parse_preamble pair
	    val ((abs_sign_entries, abs_terms), more) =	parse_body pair

	    (* Make the new module, and a temporary module. Create an 
	     * environment based upon them. *)
	    val module = Module.new_module module_name
	    val temp_mod_name = "&&&Temp&&&"
	    val temp_mod = Module.new_module temp_mod_name
	    val make_env = Module.sign_lookup
	    val env = make_env (module :: temp_mod :: (! SetUp.pervasive_mods))

	    (* Build the context created by the included modules.
	     * Merging errors can occur here *) 
	    val _ =
		app
		(fn Acc acc =>
		    (SignMerge.sign_merge module acc;
		     Module.accumulate module [acc])
	          | Imp imp =>
			(SignMerge.sign_merge module imp;
			 Module.import module [imp]))
		include_list

	    (* Thaw each abstract signature entry in the context built
	     * above, and enter it into the temporary module. Afterwards
	     * we will accumulate the temp_module with the one we are
	     * building. Fixity errors, unknown constant errors, and
	     * redefinition errors can occur here. *)
	    fun thaw ((Tp (atp, names), _) :: ls) lls =
		(store_consts temp_mod (Absyn.to_type env atp, names);
		 thaw ls lls)

	      | thaw ((Kd (akind, names), _) :: ls) lls =
		(store_types temp_mod (Absyn.to_kind akind, names);
		 thaw ls lls)

	      | thaw ((Fx arg, _) :: ls) lls =
		(update_fixities temp_mod arg;
		 thaw ls lls)

	      | thaw ((Ll (atp, names), _) :: ls) (locals, localkinds) =
		(store_consts temp_mod (Absyn.to_type env atp, names);
		 thaw ls (names @ locals, localkinds))

	      | thaw ((Lk (akind, names), _) :: ls) (l, lk) =
		(store_types temp_mod (Absyn.to_kind akind, names);
		 thaw ls (l, names @ lk))

	      | thaw nil lls = lls


	    val (locals, localkindnames) = thaw abs_sign_entries (nil, nil)

	    (* Merge the types and constants declared in the text of the
	     * module with the ones in the included modules. More merging
	     * errors can occur here. *)
	    val _ =
		SignMerge.sign_merge temp_mod module

	    (* Merging was successful, we can now accumulate the temp
	     * module into the module that we are building. *)
	    val _ =
		Module.accumulate module [temp_mod]

	    (* The temp module is no longer needed - remove it. *)
	    val _ =
		Module.remove_module temp_mod_name

	    (* Merge the module with the pervasive modules. *)
	    val _ =
		app (SignMerge.sign_merge module) (! SetUp.pervasive_mods)

	    (* Make an constant environment based upon the pervasive modules
	     * and the module which we are building. This will be used to 
	     * make the clauses in the modules. *)
	    val env =
		make_env (module :: (! SetUp.pervasive_mods))

	    (* Make terms from the abstract terms yielded by parsing in the
	     * context of the above environement, and interpret (using
	     * freevars_to_uquant) the free variables as universally
	     * quantified variables. *) 
	    val terms =
		map
		(fn (abs_M, location) =>
		 Program.freevars_to_uquant (Absyn.to_term env abs_M)
		 handle TypeRecon.TypeCheckFail (loc_opt, err_func) =>
		     Error.fatal_error
		     (Error.Location
		      (case loc_opt of
			   SOME loc => loc
			 | NONE => location,
		       Error.TypeCheckFail (err_func Print.senv))))
		abs_terms

	    (* Clausify the terms. *)
	    val clauses =
		if null terms then nil else
		Program.terms_to_clauses terms

	    (* Insert the clauses into the module. *)
	    val _ =
		Module.prog_insert module clauses

	    (* Generate a list of type constructors (localkinds) which should
	     * not appear in the type of constants in this module. This needs
	     * to be done before they are "localized" (removed from the
	     * signature below. *)
	    val localkinds =
		map (fn name =>
		     case (env name) of
			 SOME (Module.TypeCon tc) => tc
		       | _ =>
			     raise Basic.InternalError
				 ("Const with localkind name or unknown\
				  \localkind: " ^ name))
		localkindnames

	    (* Localize the local constants. *)
	    val _ =
		Module.make_local module locals

	    (* Localize the local type constructors. *)
	    val _ =
		Module.make_localkind module localkindnames

	    (* Verify that none of the localkinds appear in the types of
	     * the constants in this module. localkind_check also looks for
	     * positive occurrences of built-in types, and optionally
	     * generates a warning. *)
	    val _ =
		SignMerge.localkind_check (module, localkinds)
	in
	    (* If there is another module waiting to be read... *)
	    if more then
		parse_module stream
	    else if Error.errors_waiting () then
		(ErrorPrint.handle_errors ();
		 Basic.failure)
	    else
		(* Success!! *)
		(TerminalIO.message ("module " ^ module_name);
		 if !print_sigs then
		     Module.print_sign module
		 else
		     ();
		 Basic.success)
	end
    handle Error.Error =>
	(ErrorPrint.handle_errors ();
	 Basic.failure);
		

end (* local open Lam... *)
end; (* functor ParseModule *) 
