(* program.fun - clausification and abstraction over free variables.	*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1994,1995 by AT&T Bell Laboratories			*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Amy Felty <felty@bell-labs.com>					*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor Program 
    (structure Lg : LOGICALS
     structure Lam : LAM
     structure Error : ERROR
     structure Module : MODULE
     structure UU : UUTILS
     structure IU : IU

     sharing type Lam.lterm = Module.lterm = IU.lterm = UU.lterm
     sharing type Lg.const = Module.const = Lam.const
     sharing type IU.tp = Lg.tp = Lam.tp = Module.tp)
	 : PROGRAM =
struct

    type sterm = Lam.sterm;
    type lterm = Lam.lterm;
    type clause = Module.clause;
    type const = Lam.const;
    type tp = Lam.tp;

local
    open Lam
in

local
    open Lam.Absyn;
in

    fun abst_over_fvar ((tp, name), M) =
	let
	    fun qof (Lam (bind, M), i) =
		Lam (bind, qof (M, i + 1))
	      | qof (App (M, N), i) =
		App (qof (M, i), qof (N, i))
	      | qof (M as Fvar (_, name'), i) =
		if name = name' then
		    Index i
		else
		    M
	      | qof (Mark (lrpos, M), i) =
		qof (M, i)
	      | qof (HasType (M, _), i) =
		qof (M, i)
	      (* Const _, Index _*)
	      | qof (M, _) = M
	in
	    Lam ((tp, SOME name), qof (M, 1))
	end;

end (* local open Absyn *)

    fun freevars_to_uquant (freevars, wildvars, M) =
	let
	    val (_, M') =
		TypeRecon.type_recon_as Lg.pred_type M
	    val M'' =
		foldl
		(fn (binder as (tp, _), M) => 
		 Lam.Absyn.App
		 (Lam.Absyn.Const
		  (Lg.uquant,
		   LTerm.Type.construct_arrow
		   (LTerm.Type.construct_arrow (tp, Lg.pred_type),
		    Lg.pred_type)),
		  abst_over_fvar (binder, M)))
		M'
		(freevars @ wildvars)
	    val M''' = LTerm.sterm_to_lterm M''
	in
	    M'''
	end

    fun freevars_to_evars ((fvars, wvars, M), tp, uvars) =
	let
	    val (_, M') =
		TypeRecon.type_recon_as tp M
	    val M'' = foldl abst_over_fvar M' (fvars @ wvars)
	    val M''' = LTerm.sterm_to_lterm M''
	    val wilds =
		map (fn (tp, _) => LTerm.new_evar (tp, NONE, uvars)) wvars
	    val evars =
		map (fn (tp, name) =>
		     LTerm.new_evar (tp, SOME name, uvars))
		fvars
	    val M'''' =
		foldr (fn (e, base) =>
		       LTerm.construct_application (base, e))
		M''' (evars @ wilds)
	in
	    (evars, M'''')
	end;


    fun make_con_type tp =
	LTerm.Type.construct_arrow
	(tp, LTerm.Type.construct_arrow (tp, Lg.pred_type));

    val truth = LTerm.construct_const (Lg.truth, Lg.pred_type)

    fun gand_opt (M, N) =
	if LTerm.equal_head (Reduce.term_head M) truth then
	    N
	else if LTerm.equal_head (Reduce.term_head N) truth then
	    M
	else
	    LTerm.construct_application
	    (LTerm.construct_application
	     (LTerm.construct_const
	      (Lg.conjunction, Lam.Const.type_of Lg.conjunction), M), N)

    fun add_abst vars M =
	foldl LTerm.construct_abstraction M vars;
	
    (* For information about clausification and the following function, see
     * "A semi-functional implementation of a higher-order logic
     * programming language" by Conal Elliott and Frank Pfenning, in Peter
     * Lee, editor, Topics in Advanced Language Implementation, pages
     * 289-325, MIT Press, 1991.  *)

    fun clausify M clauses =
	let
	    exception InternalNotDForm

	    fun cy M body vars rest =
		(if UU.is_flex M then raise InternalNotDForm else ();
		 let
		     val M' = Reduce.head_normalize M
		     val ((head, tp), args) = Reduce.head_args M'
		     val (const, tp) =
			 case (LTerm.destruct_const head) of
			     SOME res => res
			   | NONE => 
				 raise Basic.InternalError
				     "clausify: Index at head"
		 in
		     cy' M ((const, tp), args) body vars rest
		 end)

	    (*
	     * The term argument of cy' should be in head normal form.
	     *)
	    and cy' const ((c, _), nil) body vars rest =
		if Lg.is_truth c then
		    rest
		else
		    Module.Clause (c, vars,
				   add_abst vars const,
				   add_abst vars body)
		    :: rest

   
	      | cy' M ((c, c_tp), [N]) body vars rest =
		if Lg.is_uquant c then
		    let
			val (binder, N') =
			    IU.theAbstraction (IU.arg_arg_type c_tp) N
		    in
			cy N' body (binder :: vars) rest
		    end

		else if Lg.is_equant c then
		    raise InternalNotDForm

		else
		    (Module.Clause (c, vars,
				    add_abst vars M,
				    add_abst vars body)) ::
		    rest

	      | cy' L ((c, c_tp), [M, N]) body vars rest =
		
		if Lg.is_conjunction c then
		    cy M body vars (cy N body vars rest)
		    		
		else if Lg.is_implication c then
		    let
			val (M', N') = Lg.std_order c (M, N)

			val body' = gand_opt (body, M')
		    in
			cy N' body' vars rest
		    end
			 
		else if Lg.is_disjunction c then
		    raise InternalNotDForm

		else
		    (Module.Clause (c, vars,
				    add_abst vars L,
				    add_abst vars body)) ::
		    rest
	
	      | cy' M ((c, _), _) body vars rest =
		(Module.Clause (c, vars, add_abst vars M, add_abst vars body))
		:: rest

	in
	    (cy M truth nil clauses)
	    handle InternalNotDForm =>
		Error.fatal_error
		(Error.Illegal
		 ("%s is not a valid D formula.",
		  [Error.STRING (Print.debug M)]))
	end
	

	fun terms_to_clauses terms =
	    let
		fun ttp nil prog = prog
		  | ttp (M :: rest) prog =
		    ttp rest (clausify M prog)
	    in
		ttp terms nil
	    end;

end (* local open Lam *)
end; (* functor Program *)
