(* query.fun - top level interface to the query engine. Defines the	*)
(* initial success continuation, and handles errors.			*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1994,1995 by AT&T Bell Laboratories			*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Amy Felty <felty@bell-labs.com>					*)
(*	Elsa Gunter <elsa@research.att.com>				*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

(* Query Loop *)

functor Query
   (structure Switch : SWITCH
    structure Lam : LAM
    structure Parse : PARSE
    structure Lg : LOGICALS
    structure Name : NAME_ENV
    structure Module : MODULE
    structure IU : IU
    structure Program : PROGRAM
    structure Error : ERROR
    structure ErrorPrint : ERROR_PRINT
    structure Interpreter : INTERPRETER
    structure Interface : PARSER_INTERFACE
    structure TerminalIO : TERMINAL_IO
    structure SetUp : SET_UP

    exception exn of Lam.lterm

    sharing type Lam.Absyn.aterm = Interface.aterm
    sharing type Module.const = Lam.const = IU.const
    sharing type Lam.Absyn.sign_entry = Module.sign_entry
    sharing type Interface.parse_result = Parse.parse_result
    sharing type Error.location = Parse.location = Lam.TypeRecon.location
    sharing type IU.i_record = Interpreter.i_record
    sharing type IU.env = Lam.Print.name_env = Name.name_env =
	Lam.TypeRecon.name_env
    sharing type IU.module = Module.module = SetUp.module
    sharing type Lam.lterm = Interpreter.lterm = Program.lterm
    sharing type Lam.constraint = (* Constraints.constraint =*)
	Interpreter.constraint
    sharing type Lam.tp = Lg.tp = Program.tp
    sharing type Program.sterm = Lam.sterm)
	: QUERY =
struct


    (* Switches *)
    val all_solutions = ref false
    val _ = Switch.make "all_solutions" (Switch.make_on_off all_solutions);

    val primary_prompt = ref "?- ";
    val secondary_prompt = ref " - ";
    val _ = Switch.make "primary_query_prompt" (fn x => primary_prompt := x,
						fn () => !primary_prompt)
    val _ = Switch.make "secondary_query_prompt" (fn x =>secondary_prompt := x,
						  fn () => !secondary_prompt)
    (* End Switches *)

local
    open Error Lam
in

    type module = Module.module;
    type sterm = LTerm.sterm;
    type lterm = Lam.lterm;
    type const = Lam.const;
    type tp = Program.tp;
    type parse_stream = Parse.parse_stream


    exception EOF;
    exception Query_Complete

    fun query_with_env name_env module raw_term =
	(let
	     val _ = Const.reset_base_time ();

	     (* the uvars argument to freevars_to_evars is nil because the
	      * free variables in the query term are quantified at the
	      * outermost level. *)
	     val (evars, M) =
		 Program.freevars_to_evars (raw_term, Lg.pred_type, nil)
		 handle TypeRecon.TypeCheckFail (loc_opt, str_func) =>
		     fatal_error
		     (case loc_opt of
			  NONE =>
			      TypeCheckFail (str_func name_env)
			| SOME loc => 
			      Location (loc, TypeCheckFail(str_func name_env)))

	     (* This is a small hack: probably the absyn should return the
	      * freevars in the order in which they are found. *)
	     val evars = Lib.List.reverse evars
		 
	    (* This mapping serves to "register" all of the evar names in
	     * the naming environment. This will ensure that all evars to
	     * which they are bound will be displaced as one of these
	     * evars, rather than the other way around.  *)
	     val _ = map (Print.makestring_lterm name_env) evars;

	     (* Notice that the local variables are created *after* the evars
	      * in the query term, i.e. the free variables in a query are
	      * quantified at the *outermost* level. *)
	     val uvars = Module.inst_locals module

	     val at_least_one = ref false

	     (* The top level ctag exception - used to implement ! *)
	     exception top_ctag

	     (* The initial success continuation *)
	     fun sc newcon =
		 let
		     val newcon' = Simplify.simplify_constraint newcon
		 in
		     (if null evars then
			  (TerminalIO.print "\nsolved\n";
			   raise Query_Complete)
		      else
			  TerminalIO.print
			  (Print.makestring_substitution name_env evars);

		      if Constraints.is_empty_constraint newcon' then
			  ()
		      else 
			  TerminalIO.print
			  ("\n" ^
			   (Print.makestring_constraint name_env newcon'));

		      if (! all_solutions) orelse
			  ((Lib.String.ordof
			    (Lib.IO.input_line (! TerminalIO.std_in), 0))
			   = Lib.String.ordof (";", 0))
			  (* input_line returns the empty string for ^D,
			   causing Subscript to be raised *)
			  handle Lib.String.Subscript => false then
			      at_least_one := true
		      else
			  (TerminalIO.print "\n";
			   raise Query_Complete))
		 end
	 in
	     (at_least_one := false;

	      (* Call the interpreter on the new query with the initial
	       * success continuation. *)		     

	      (Interpreter.solve (M, IU.IR
				  {modules = module :: (!SetUp.pervasive_mods),
				   env = name_env, ctag = top_ctag,
				   traced = [], clauses = [], uvars = uvars}, 
				  Constraints.empty_constraint,
				  sc))
	      handle top_ctag => ();
		  
	      (* If we make it this far, then we have failed either
	       * to find more solutions upon request, or to find the
	       * first solution. *)
	      TerminalIO.print (if !at_least_one then
				    "\nno more solutions\n"
				else
				    "\nno\n"))
	 end) handle Query_Complete => ()

    fun handle_exn env excp =
	TerminalIO.message
	("Error: Uncaught exception "^ (Print.makestring_lterm env excp));

    fun query module raw_term =
	let
	    val env =
		Name.make_env
		(Module.is_const (module :: (! SetUp.pervasive_mods)))
	in
	    (query_with_env env module raw_term;
	     Basic.success)
	    handle
	    Interpreter.Quit => Basic.success
	  | exn excp =>
		(handle_exn env excp;
		 Basic.failure)
	  | Error.Error => 
		(ErrorPrint.handle_errors ();
		 Basic.failure)
	end;

    fun interactive_query stream module =
	let
	    val modules = module :: (! SetUp.pervasive_mods)
	    val env = Module.sign_lookup modules

	    fun iquery () =
		let
		    val name_env =
			Name.make_env (Module.is_const modules)

		    val (raw_term, location) =
			(case Parse.parse
			     (!primary_prompt, !secondary_prompt) stream of
			     (SOME (Interface.ParsedLp
				    (Interface.Term result), location)) =>
				 (Absyn.to_term env result, location)
			   | (SOME (Interface.ParsedScript Interface.End,_)) =>
				 raise EOF
 			   | (SOME (_, location)) =>
				 fatal_error
				 (Location
				  (location,
				   (Illegal
				    ("Illegal statement.", []))))
			   | NONE => 
				 raise EOF)
		in
		    (query_with_env name_env module raw_term)
		    handle
		    exn excp =>
			handle_exn name_env excp
		end
	    handle
	    Error.Error =>
		ErrorPrint.handle_errors ()
	
	    fun loop () =
		(iquery ();
		 loop ())
		handle 
		Interpreter.Quit => Basic.success

	      (* should be handled by failure?? *)
	      | EOF => Basic.success

    in
	loop ()
    end;

end; (* local *)
end; (* functor Query *)

