(* script.fun - Terzo scripting language.				*)

(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Author:								*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor Script
    (structure Error : ERROR
     structure ErrorPrint : ERROR_PRINT
     structure TerminalIO : TERMINAL_IO
     structure Module : MODULE
     structure SignMerge : SIGN_MERGE
     structure Absyn : ABSYN
     structure Query : QUERY
     structure Interface : PARSER_INTERFACE
     structure Parse : PARSE
     structure ParseModule : PARSE_MODULE
     structure Switch : SWITCH
     structure Help : HELP

     val pervasive_mods : Module.module list ref
     val boot_path : string list

     sharing type Module.sign_entry = Absyn.sign_entry
     sharing type Error.error = ErrorPrint.error
     sharing type Error.location = Parse.location
     sharing type Query.tp = Absyn.tp
     sharing type Absyn.aterm = Interface.aterm
     sharing type Query.sterm = Absyn.sterm
     sharing type Query.module = Module.module = SignMerge.module
     sharing type Interface.parse_result = Parse.parse_result
     sharing type Parse.parse_stream = ParseModule.parse_stream =
	 Query.parse_stream)
	 : SCRIPT =
struct

local
    open Interface Error
in

    (* Switches *)
    val path = ref boot_path;

    val primary_prompt = ref "Terzo> ";
    val secondary_prompt = ref "     > ";

    val _ = Switch.make "primary_loader_prompt"
	(fn x => primary_prompt := x, fn () => !primary_prompt)
    val _ = Switch.make "secondary_loader_prompt"
	(fn x => secondary_prompt := x, fn () => !secondary_prompt)

    val _ = Switch.make "on"
	(fn _ =>
	 error (Illegal ("The on switch is not set-able.", [])),
	 fn () => "on")

    val _ = Switch.make "off"
	(fn _ =>
	 error (Illegal ("The off switch is not set-able.", [])),
	 fn () => "off")
    
local
    fun set_path new_path = 
	path := (Switch.explode new_path);
	
    fun show_path () =
	Switch.implode (! path);
in
    val _ = Switch.make "path" (set_path, show_path)
end
(* end of switches *)


    type parse_stream = Parse.parse_stream;

    (* read_script reads a Terzo script. If an error is signaled in the process
     * of reading that script, propagates the error. *)
    fun read_script stream =
	let
	    fun parse () =
		(Parse.parse (!primary_prompt, !secondary_prompt) stream)

	    fun get_module location module_name =
		case (Module.lookup_module module_name) of
		    SOME module => module
		  | NONE => 
			fatal_error
			(Location (location, UnknownModule module_name))

	    fun acc_modules location module_names =
		let
		    val modules = map (get_module location) module_names
		    val acc_mod_name =  "&&&ACC_MOD&&&"
		    val acc_mod = Module.new_module acc_mod_name
		in
		    (app (fn acc =>
			  (SignMerge.sign_merge acc_mod acc;
			   Module.accumulate acc_mod [acc]))
		     modules;
		     Module.remove_module acc_mod_name;
		     acc_mod)
		end

	    fun parse_script (SOME (ParsedScript (Load files), location)) =
		let
		    fun while_successful func (l::ls) =
			if (Basic.successful (func l)) then
			    while_successful func ls
			else
			    Basic.failure
		      | while_successful _ nil = Basic.success;

		    fun read_file file =
			let
			    val file_and_path =
				if Lib.Dir.readable file then
				    file
				else
				    (Lib.Dir.find_file (!path, file))
				    handle Lib.Dir.NoSuchFile =>
					fatal_error
					(Location
					 (location,
					  Illegal ("File %s not found.",
						   [STRING file])))

			    val (close_func, new_stream) =
				Parse.file_init file_and_path

			    val initial =
				Parse.parse ("", "") new_stream

			    val _ = Parse.unparse new_stream initial
			in
			    (fn x => (close_func (); x))
			    (case initial of
				 SOME (ParsedModule (Module _), _) =>
				     ParseModule.parse_module new_stream
			       | _ => 
				     read_script new_stream)
			end
		in
		    if Basic.successful (while_successful read_file files) then
			parse_script (parse ())
		    else
			raise Error
		end
		
	      | parse_script (SOME (ParsedScript (Set (switch, value)),
				    location))=
		((let
		      fun strip_list ((Val v)::ls) result =
			  strip_list ls (result ^ v)
			| strip_list ((Var v)::ls) result =
			  strip_list ls (result ^ (Switch.show v))
			| strip_list [] result = result
		  in
		      Switch.set switch (strip_list value "")
		  end);
		 parse_script (parse ()))
		  
	      | parse_script (SOME (ParsedScript (Show NONE), _)) =
		(Switch.show_all
		 (fn (name, value) =>
		  TerminalIO.message (Switch.makestring_switch (name, value)));
		 parse_script (parse ()))
	
	      | parse_script (SOME (ParsedScript (Show (SOME switch)),
				    location)) =
		((TerminalIO.message 
		  (Switch.makestring_switch (switch, (Switch.show switch))));
		 parse_script (parse ()))

	      | parse_script (SOME
			      (ParsedScript (Print module_name), location)) =
		(Module.print_sign (get_module location module_name);
		 parse_script (parse ()))

	      | parse_script (SOME (ParsedScript (Help NONE), _)) =
		(Help.general ();
		 parse_script (parse ()))
		
	      | parse_script (SOME (ParsedScript (Help (SOME topic)), _)) =
		(Help.lookup topic;
		 parse_script (parse ()))

	      | parse_script (SOME (ParsedScript Begin, location)) =
		(ParseModule.parse_module stream;
		 parse_script (parse ()))

	      | parse_script (SOME (ParsedScript End, location)) =
		fatal_error
		(Location (location,
			   Illegal ("Found #end without matching begin.", [])))

	      | parse_script
		(SOME (ParsedScript (Query (SOME (module_list, NONE))),
		       location)) =
		(Query.interactive_query stream
		 (acc_modules location module_list);
		 parse_script (parse ()))

	      | parse_script
		(SOME (ParsedScript (Query (SOME (module_list,SOME abs_term))),
		       location)) =
		let
		    val module = acc_modules location module_list
		in
		    (Query.query module
		     (Absyn.to_term
		      (Module.sign_lookup (module :: (! pervasive_mods)))
		      abs_term);
		     parse_script (parse ()))
		end

	      | parse_script (SOME (ParsedScript (Query NONE), location)) =
		(Query.interactive_query stream (get_module location "System");
		 parse_script (parse ()))

	      | parse_script NONE = ()
		
	      | parse_script (SOME (ParsedScript Quit, _)) = ()
		
	      | parse_script (SOME (_, location)) =
		fatal_error
		(Location
		 (location,
		  Illegal ("Illegal declaration: not parsing module.", [])))

	in
	    (parse_script (parse ());
	     
	     if Error.errors_waiting () then
		 (Error.handle_errors ErrorPrint.error_print;
		  Basic.failure)
	     else
		 Basic.success)
	end handle
	Error.Error =>
	    (Error.handle_errors ErrorPrint.error_print;
	     Basic.failure);

end; (* local open Interface Error *)
end; (* functor Script *)
