(* parse_core.fun - parser primitives					*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)

functor ParseCore
  (structure Location : LOCATION
   structure TerminalIO : TERMINAL_IO
   structure Module : MODULE
   structure Error : ERROR
   structure ParserInterface : PARSER_INTERFACE
   structure Parser : ARG_PARSER

   exception Error of (int * int) * Error.error
   datatype parserarg =
       A of {counter : Location.counter,
	     name : string}
   val EOF_token_name : Parser.pos * Parser.pos
       -> (Parser.svalue,Parser.pos) Parser.Token.token
   val EOP_token_name : Parser.pos * Parser.pos
       -> (Parser.svalue,Parser.pos) Parser.Token.token

   sharing type Parser.lexarg = Location.lexarg
   sharing type Parser.arg = parserarg
   sharing type Parser.pos = Location.pos
   sharing type Parser.result = ParserInterface.result 
   sharing type Location.location = Error.location = ParserInterface.location)
       : PARSE =
struct

    structure Parser = Parser;

    type parse_result= ParserInterface.parse_result;
    type location = Location.location;

val EOF_token = EOF_token_name (Location.empty_pos, Location.empty_pos)
val EOP_token = EOP_token_name (Location.empty_pos, Location.empty_pos)

type token_stream =
    (Parser.svalue, Parser.pos) Parser.Token.token Parser.Stream.stream;

datatype stream_type =
    Interactive of bool ref * (string * string) ref * Lib.IO.instream
  | NonInteractive;



datatype parse_stream = 
    PS of
    {name : string,
     counter : Location.counter,
     stream_type : stream_type,
     token_stream : token_stream ref,
     stack : (parse_result * Location.location) option list ref} ref;

fun stream_name (PS (ref {name = name, ...})) = name;


fun is_all_whitespace str =
    let
	fun iaw i =
	    if not (Lib.String.is_whitespace
		    (Lib.String.substring (str, i, 1))) then
		false
	    else
		iaw (i + 1)
    in
	iaw 0 handle Ord => true
    end;


fun interactive_stream_init in_stream name =
    let
	val first_time = ref true
	val counter = Location.new_counter ()
	val prompts_ref = ref ("not set 1 ", "not set 2 ")
	    
	fun input_fun _ =
	    (if !first_time then
		 TerminalIO.print (#1 (! prompts_ref))
	     else
		 TerminalIO.print (#2 (! prompts_ref));
	     let
		 val line = Lib.IO.input_line in_stream
	     in
		 (if is_all_whitespace line then () else first_time := false;
		  line)
	     end)
		     
	val token_stream =
	    Parser.makeLexer input_fun (Location.LA (name, counter))
    in
	PS (ref {name = name, counter = counter, stack = ref [],
		 token_stream = ref token_stream, stream_type =
		 Interactive (first_time, prompts_ref, in_stream)})
    end;


fun reinit_stream
    (PS (s_ref as (ref {name = name,
			stream_type = Interactive (_, _, in_stream), ...}))) =
    let
	val PS (ref s) = interactive_stream_init in_stream name
    in
	(s_ref := s;
	 PS s_ref)
    end

  (* Reinit a non-interactive stream. *)
  | reinit_stream
    (ps as PS (ref {name = name, counter = counter,
		    token_stream = token_stream, ...})) =
    
    let
	fun clear_line (token, new_stream) =
	    if Parser.sameToken (token, EOF_token) orelse
	       Parser.sameToken (token, EOP_token) then
	       new_stream
	    else
		clear_line (Parser.Stream.get new_stream)
    in
	(token_stream := clear_line (Parser.Stream.get (! token_stream));
	 ps)
    end;

fun stream_init in_stream name =
    let
	val counter = Location.new_counter ()
	val stream =
	    Parser.makeLexer (fn n => Lib.IO.input (in_stream, n))
	    (Location.LA (name, counter))
    in
	PS (ref {name = name, counter = counter, stream_type = NonInteractive,
		 token_stream = ref stream, stack = ref []})
    end;

fun file_init filename =
   let
       val in_stream = Lib.IO.open_in filename
       fun close_func () =
	   (Lib.IO.close_in in_stream;
	    TerminalIO.message ("[closed file " ^ filename ^ "]"))
   in
       (TerminalIO.message ("[reading file " ^ filename ^ "]");
	(close_func, stream_init in_stream filename))
   end;

exception ParseError of (int * int) * string;
fun parsing_error (msg, lpos, rpos) =
    raise ParseError ((lpos, rpos), msg);

fun parse _ (PS (ref {stack = stack_ref as ref (hd::tl), ...})) =
    (stack_ref := tl;
     hd)
    
  | parse prompts
    (stream as 
     PS (ref {name, counter, stream_type, token_stream, ...})) =
    (case stream_type of
	 Interactive (first_time, prompts_ref, _) =>
	     (first_time := true;
	      prompts_ref := prompts)
       | NonInteractive => ();
     let
	 val (next_token, stream1) = Parser.Stream.get (! token_stream)
     in
	 if Parser.sameToken (next_token, EOF_token) then
	     NONE
	 else
	     let
		 val (ParserInterface.PR (parsed_item, location), stream1) =
		     Parser.parse (0, ! token_stream, parsing_error,
				   A {counter = counter,
				      name = name})
		     

		 (* pwerr *)
		 (* skip the terminating token *)
		 val (_, stream2) = Parser.Stream.get stream1


	     in
		 (* Update the stream ref *)
		 (token_stream := stream2;
		  
		  (* return object parsed and location *)
		  SOME (parsed_item, location))
	     end
     end)
	 handle
	 ParseError (lrpos, msg) =>
	     (reinit_stream stream;
	      Error.fatal_error
	      (Error.Location
	       (Location.make_location counter name lrpos,
		Error.Parse (msg, []))))
       | Error.Error =>
	     (reinit_stream stream;
	      raise Error.Error)


fun is_interactive (PS (ref {stream_type = Interactive _, ...})) =
    true
  | is_interactive _ =
    false;

fun unparse (PS (ref {stack = stack, ...})) result_location =
    stack := result_location :: (! stack);

end  (* functor ParseCore *)
