(* lib.fun - basic library						*)

(* Copyright (c) 1991,1994,1995 by Carnegie Mellon University		*)
(* Copyright (c) 1995,1996 by the University of Pennsylvania		*)

(* For information about the distributaion, copying, and modification	*)
(* of this software, please read the file COPYING located in the root	*)
(* directory of this distribnutaion. If you did not receive the file	*)
(* COPYING write to Philip Wickline at the address below.		*)

(* This program is distributed in the hope that it will be useful but	*)
(* WITHOUT ANY WARRANTY; without even the implied warranty of		*)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file	*)
(* COPYING for more details.						*)

(* For information about the structure of the Terzo lambdaProlog	*)
(* implementation, see the file src/README in this distribution. Please	*)
(* address any questions about this code or the Terzo lambdaProlog	*)
(* implementation to Philip Wickline at <philipw@saul.cis.upenn.edu>.	*)

(* Authors:								*)
(*	Frank Pfenning <fp@cs.cmu.edu>					*)
(*	Philip Wickline <philipw@saul.cis.upenn.edu>			*)


structure Lib : LIB =
    struct

	structure Arith =
	struct
	    exception Overflow = Overflow;
	    exception Div = Div;
	end;

	structure Convert =
	struct
	    exception Convert;

	    fun atoi s =
		(case (Int.fromString s) of
		     NONE => raise Convert
		   | SOME i => i)
		     handle Overflow => raise Convert;

	    fun atof s =
		(case (Real.fromString s) of
		     NONE => raise Convert
		   | SOME i => i)
		     handle Overflow => raise Convert;

	    val itoa = Int.toString;
	    val ftoa = Real.toString;
	    val btoa = Bool.toString;
	end;
	    
	structure Sys =
	struct
	    fun date () = 
		Date.toString (Date.fromTimeLocal (Time.now ()));
	end;

	structure IO =
        struct
	    exception Io of string;

	    type instream = TextIO.instream;
	    type outstream = TextIO.outstream;

	    val std_in = TextIO.stdIn;
	    val std_out = TextIO.stdOut;
	    val std_err = TextIO.stdErr;

	    val open_string = TextIO.openString;
	    val open_in = TextIO.openIn;
	    val open_out = TextIO.openOut;
	    val open_append = TextIO.openAppend;

	    val close_in = TextIO.closeIn;
	    val close_out = TextIO.closeOut;

	    val flush_out = TextIO.flushOut;
	    val input = TextIO.inputN;
	    fun output (out, vec) =
		(TextIO.output (out, vec);
		 TextIO.flushOut out);
	    fun lookahead instream =
		(case (TextIO.lookahead instream) of
		     NONE => ""
		   | SOME c => String.str c);
	    val input_line = TextIO.inputLine;
	    val end_of_stream = TextIO.endOfStream;

	end;

	structure List =
	    struct
		exception Subscript = Subscript;

		fun lookup eq obj ((h, t) :: ls) =
		    if eq (obj, h) then
			SOME t
		    else
			lookup eq obj ls
		  | lookup _ _ _ = NONE

		fun flatten nil = nil
		  | flatten (l::ls) =
		    l @ (flatten ls);

		fun filter _ nil = nil
		  | filter f (x::xs) =
		    (case (f x) of
			 NONE => filter f xs
		       | SOME x' => x' :: (filter f xs));

		fun reverse ls =
		    let
			fun rh nil out = out
			  | rh (l::ls) out = rh ls (l::out)
		    in
			rh ls []
		    end;
		val nth = List.nth;
		val foldl = foldl;
		val foldr = foldr;
		val exists = List.exists;

		fun merge cmp (ls1, nil) = ls1
		  | merge cmp (nil, ls2) = ls2
		  | merge cmp (x1::ls1, x2::ls2) =
		    (if cmp (x1, x2) then x1 else x2) :: (merge cmp (ls1, ls2));

		fun split (ls1, nil, _) = (List.rev ls1, nil)
		  | split (ls1, ls2, 0) = (List.rev ls1, ls2)
		  | split (ls1, x::ls2, i) = split (x::ls1, ls2, i - 1);

		fun sort cmp nil = nil
		  | sort cmp ls =
		    let
			val (hd, tl) = split (nil, ls, length ls div 2)
		    in
			merge cmp (sort cmp hd, sort cmp tl)
		    end;

	    end;

	structure Signal =
	struct
	    exception Return

	    fun handle_interrupt (loop : unit -> unit) =
	     let
		    val oldHandler =
			Signals.inqHandler Signals.sigINT
		    fun clean_up () =
			(Signals.setHandler (Signals.sigINT, oldHandler);
			 ())
		in
		    (SMLofNJ.Cont.callcc
		     (fn k =>
		      (Signals.setHandler
		       (Signals.sigINT,
			Signals.HANDLER (fn _ => (print "\ninterrupt\n"; k)));
		       loop ()));

		     loop ();
		     clean_up ())
		    handle Return =>
			clean_up ()
			 | exn =>
			       (clean_up ();
				raise exn)
		end

	end;

	structure Misc =
	struct
	    fun any func (l :: ls) =
		(func l) orelse (any func ls)
	      | any _ nil = false;
	    fun uncurry func (a, b) = func a b;
	    fun curry func a b = func (a, b);
	    fun equal a b = a = b;
	    fun pre_inc i_ref =
		(i_ref := (! i_ref) + 1;
		 ! i_ref);
	    fun pre_dec i_ref =
		(i_ref := (! i_ref) - 1;
		 ! i_ref);
	    fun post_inc (i_ref as (ref i)) =
		(i_ref := (! i_ref) + 1;
		 i);
	    fun post_dec (i_ref as (ref i)) =
		(i_ref := (! i_ref) - 1;
		 i);
	    fun id x = x;
	end;

	structure Int =
	struct
	    open Int;
	    val hash_int = Word.fromInt;
	end;

	structure Real = Real;

	structure Math = Math;
		
	structure String =
        struct
	    exception Subscript = Subscript;
	    exception Chr = Chr;

	    fun hash_string s = foldl (fn (c, h) => h + 123 + (Char.ord c)) 0 (String.explode s)

	    fun tokenize separator string =
		let
		    fun tk (nil, sout) =
			if sout <> "" then
			    [sout]
			else
			    []
		      | tk (s::ss, sout) =
			if s = separator then
			    if sout <> "" then
				sout :: (tk (ss, ""))
			    else
				tk (ss, "")
			else
			    tk (ss, sout ^ s)
		in
		    tk (map String.str (String.explode string), "")
		end;

	    fun up c =
		if Char.isLower c then
		    chr ((ord c) - 32)
		else
		    c

	    fun down c =
		if Char.isUpper c then
		    chr ((ord c) + 32)
		else
		    c

	    fun upcase str =
		String.implode (map up (String.explode str));

	    fun downcase str =
		String.implode (map down (String.explode str));

	    val implode = String.concat;

	    fun explode s = map String.str (String.explode s);

	    val size = String.size;

	    val lessthan = String.< ;

	    val substring = String.substring;

	    fun is_whitespace str =
		Char.isSpace
		(case (Char.fromString str) of
		     NONE => raise Subscript
		   | SOME b => b);

	    val chr = String.str o Char.chr;

	    val ordof = Char.ord o String.sub;
	end;

	structure Env =
	struct
	    fun get_value s = OS.Process.getEnv s;
	end;

	structure Process = OS.Process;

	structure Dir =
	struct
	    exception NoSuchFile;

	    val cwd = OS.FileSys.getDir;

	    fun mk_path (dir, file) =
		OS.Path.joinDirFile {dir = dir, file = file};

	    fun readable file =
		OS.FileSys.access (file, [OS.FileSys.A_READ])
		(* This is to fix a bug in sml/nj it shouldn't really have to be here *)
		handle SysErr => false;

	    (* Returns the full path of filename if it exists, and is
	     * readable, in the current path, and signals an error
	     * otherwise. *)
	    fun find_file (nil, _) =
		raise NoSuchFile
	      | find_file (l::ls, filename) =
		let
		    val path_file = mk_path (l, filename)
		in
		    if readable path_file then
			path_file
		    else
			find_file (ls, filename)
		end;

	end;

	structure Table : TABLE = TableHash ();

    end;
